import numpy as np

from .._io.input._helpers import read
from ._cylindric_grid import cylindric_grid
from ._structured_grid import structured_grid


def from_meshmaker(filename_or_dict):
    """
    Generate a mesh from a block MESHM.

    Parameters
    ----------
    filename_or_dict: str or dict
        Input file name or parameters dict with key "meshmaker".

    """
    if isinstance(filename_or_dict, str):
        parameters = read(filename_or_dict, file_format="tough")

    else:
        parameters = filename_or_dict

    if "meshmaker" not in parameters:
        raise ValueError()

    if "type" not in parameters["meshmaker"]:
        raise ValueError()

    if parameters["meshmaker"]["type"] not in {"xyz", "rz2d", "rz2dl"}:
        raise ValueError()

    # XYZ
    if parameters["meshmaker"]["type"] == "xyz":
        dx_, dy_, dz_ = parse_xyz(parameters["meshmaker"]["parameters"])

        dx, dy, dz = [], [], []
        for increment in dx_:
            append(dx, **increment)
        for increment in dy_:
            append(dy, **increment)
        for increment in dz_:
            append(dz, **increment)

        if not len(dx):
            dx = [1.0]
        if not len(dy):
            dy = [1.0]
        if not len(dz):
            dz = [1.0]

        return structured_grid(dx, dy, dz)

    # RZ2D
    else:
        dr_, dz_ = parse_rz2d(parameters["meshmaker"]["parameters"])

        dr, dz = [], []
        for increment in dr_:
            append(dr, **increment)
        for increment in dz_:
            append(dz, **increment)

        if not len(dr):
            dr = [1.0]
        if not len(dz):
            dz = [1.0]

        return cylindric_grid(dr, dz, layer=parameters["meshmaker"]["type"] == "rz2dl")


def parse_xyz(parameters):
    """Parse input for XYZ mesh."""

    def parse(parameter):
        n = get_value(parameter, "n_increment", 1)
        if n <= 0:
            raise ValueError()

        sizes = get_value(parameter, "sizes", 0.0)
        if np.ndim(sizes) == 0:
            if sizes < 0.0:
                raise ValueError()

            tmp = [
                {
                    "n_increment": n,
                    "size": sizes,
                    "type": "uniform",
                    "radius_ref": None,
                }
            ]

        elif np.ndim(sizes) == 1:
            if (np.array(sizes) < 0.0).any():
                raise ValueError()

            tmp = [
                {
                    "n_increment": count,
                    "size": value,
                    "type": "uniform",
                    "radius_ref": None,
                }
                for count, value in zip(*squeeze(sizes))
            ]

        return tmp

    dx, dy, dz = [], [], []
    for parameter in parameters:
        if parameter["type"] == "nx":
            dx += parse(parameter)

        elif parameter["type"] == "ny":
            dy += parse(parameter)

        elif parameter["type"] == "nz":
            dz += parse(parameter)

    return dx, dy, dz


def parse_rz2d(parameters):
    """Parse input for RZ2D mesh."""
    rmax = 0.0
    dr, dz = [], []
    for parameter in parameters:
        if parameter["type"] == "radii":
            if len(parameter["radii"]) < 2:
                continue

            tmp = np.diff(parameter["radii"])
            if (tmp < 0.0).any():
                raise ValueError("radii must be sorted in ascending order.")

            rmax += tmp.sum()
            dr += [
                {
                    "n_increment": count,
                    "size": value,
                    "type": "uniform",
                    "radius_ref": None,
                }
                for count, value in zip(*squeeze(tmp))
            ]

        elif parameter["type"] == "equid":
            n = get_value(parameter, "n_increment", 1)
            if n <= 0:
                raise ValueError()

            size = get_value(parameter, "size", 0.0)
            if size < 0.0:
                raise ValueError()

            rmax += n * size
            dr.append(
                {"n_increment": n, "size": size, "type": "uniform", "radius_ref": None,}
            )

        elif parameter["type"] == "logar":
            n = get_value(parameter, "n_increment", 1)
            if n <= 0:
                raise ValueError()

            radius = get_value(parameter, "radius", 0.0)
            if radius <= rmax:
                raise ValueError()

            else:
                size = radius - rmax

            radius_ref = get_value(parameter, "radius_ref", 0.0)
            if radius_ref < 0.0:
                raise ValueError()

            rmax += size
            dr.append(
                {
                    "n_increment": n,
                    "size": size,
                    "type": "logarithmic",
                    "radius_ref": radius_ref,
                }
            )

        elif parameter["type"] == "layer":
            dz += [
                {
                    "n_increment": count,
                    "size": value,
                    "type": "uniform",
                    "radius_ref": None,
                }
                for count, value in zip(*squeeze(parameter["thicknesses"]))
            ]

    return dr, dz


def get_value(parameter, key, default):
    """Helper function to get value in dictionary."""
    return (
        default
        if key not in parameter
        or (isinstance(parameter[key], list) and not len(parameter[key]))
        or parameter[key] is None
        else parameter[key]
    )


def squeeze(data):
    """Squeeze a sequence."""
    count = [1]
    values = [data[0]]

    for value in data[1:]:
        if value == values[-1]:
            count[-1] += 1

        else:
            values.append(value)
            count.append(1)

    return count, values


def append(sizes, n_increment, size, type="uniform", radius_ref=None):
    """Append next increment size(s)."""
    if type == "uniform":
        sizes += [size] * n_increment

    elif type == "logarithmic":
        if not radius_ref:
            if not len(sizes):
                raise ValueError()

            radius_ref = sizes[-1]

        f = get_factor(n_increment, size, radius_ref)

        sizes.append(f * radius_ref)
        for _ in range(n_increment - 1):
            sizes.append(f * sizes[-1])


def get_factor(n_increment, radius, radius_ref):
    """
    Calculate f factor.

    Note
    ----
    Solve the equation DR * f + DR * f ** 2 + ... + DR * f ** n = RMAX.
    This polynomial should have only one positive root.

    """
    roots = np.roots(np.append(np.ones(n_increment), -radius / radius_ref))
    roots = roots[::-1]  # The real positive root seems to be the last one

    # Still check that it's real and positive
    f = None
    for root in roots:
        if np.isreal(root) and root > 0.0:
            f = root.real
            break

    if f is None:
        raise ValueError("no real positive root found.")

    return f



from __future__ import with_statement

import numpy as np

from ...._common import open_file
from .._common import to_output

__all__ = [
    "read",
    "write",
]

header_to_unit = {
    "ELEM": "",
    "X": "M",
    "Y": "M",
    "Z": "M",
    "PRES": "PA",
    "P": "PA",
    "TEMP": "DEC-C",
    "T": "DEC-C",
    "PCAP_GL": "PA",
    "PCAP": "PA",
    "DEN_G": "KG/M**3",
    "DG": "KG/M**3",
    "DEN_L": "KG/M**3",
    "DW": "KG/M**3",
    "ELEM1": "",
    "ELEM2": "",
    "HEAT": "W",
    "FLOW": "KG/S",
    "FLOW_G": "KG/S",
    "FLOW_L": "KG/S",
}


def read(filename, file_type, file_format, labels_order):
    """Read OUTPUT_{ELEME, CONNE}.csv."""
    with open_file(filename, "r") as f:
        headers, times, variables = _read_csv(f, file_type)

        ilab = 1 if file_type == "element" else 2
        headers = headers[ilab:]
        labels = [[v[:ilab] for v in variable] for variable in variables]
        labels = (
            [[l[0] for l in label] for label in labels]
            if file_type == "element"
            else labels
        )
        variables = np.array([[v[ilab:] for v in variable] for variable in variables])

    return to_output(
        file_type, file_format, labels_order, headers, times, labels, variables
    )


def _read_csv(f, file_type):
    """Read CSV table."""
    # Read header
    line = f.readline().replace('"', "")
    headers = [l.strip() for l in line.split(",")]

    # Skip second line (unit)
    line = f.readline()

    # Check third line (does it start with TIME?)
    line = f.readline()
    single = not line.startswith('"TIME')

    # Read data
    if single:
        times, variables = [None], [[]]
    else:
        times, variables = [], []

    line = line.replace('"', "").strip()
    ilab = 1 if file_type == "element" else 2
    while line:
        line = line.split(",")

        # Time step
        if line[0].startswith("TIME"):
            line = line[0].split()
            times.append(float(line[-1]))
            variables.append([])

        # Output
        else:
            tmp = [l.strip() for l in line[:ilab]]
            tmp += [float(l.strip()) for l in line[ilab:]]
            variables[-1].append(tmp)

        line = f.readline().strip().replace('"', "")

    return headers, times, variables


def write(filename, output, unit=None):
    """Write OUTPUT_{ELEME, CONNE}.csv."""
    if not (unit is None or isinstance(unit, dict)):
        raise TypeError()

    out = output[-1]
    headers = ["ELEM"] if out.type == "element" else ["ELEM1", "ELEM2"]
    headers += ["X"] if "X" in out.data.keys() else []
    headers += ["Y"] if "Y" in out.data.keys() else []
    headers += ["Z"] if "Z" in out.data.keys() else []
    headers += [k for k in out.data.keys() if k not in {"X", "Y", "Z"}]

    with open_file(filename, "w") as f:
        _write_csv(f, output, headers, unit)


def _write_csv(f, output, headers, unit=None):
    """Write CSV table."""
    header_to_unit_ = header_to_unit.copy()
    if unit is not None:
        header_to_unit_.update(unit)

    # Headers
    units = [
        header_to_unit_[header] if header in header_to_unit_.keys() else "-"
        for header in headers
    ]
    f.write(",".join('"{:>18}"'.format(header) for header in headers) + "\n")
    f.write(",".join('"{:>18}"'.format("({})".format(unit)) for unit in units) + "\n")

    # Data
    formats = [
        '"{:>18}"' if header.startswith("ELEM") else "{:20.12e}" for header in headers
    ]
    for out in output:
        # Time step
        f.write('"TIME [sec]  {:.8e}"\n'.format(out.time))

        # Table
        for i, label in enumerate(out.labels):
            record = [label] if isinstance(label, str) else [l for l in label]
            record += [out.data[k][i] for k in headers if not k.startswith("ELEM")]
            record = (
                ",".join(
                    fmt.format(rec) if rec is not None else fmt.format(0.0)
                    for fmt, rec in zip(formats, record)
                )
                + "\n"
            )
            f.write(record)



from __future__ import division, unicode_literals, with_statement

import logging
import os

import numpy as np

from ..._common import open_file
from ._helpers import block

__all__ = [
    "read",
    "write",
]


def read(filename, label_length=None):
    """Read TOUGH MESH file."""
    from ... import read_input

    mesh = read_input(filename, file_format="tough", label_length=label_length)
    return {
        k: v
        for k, v in mesh.items()
        if k
        in {
            "elements",
            "elements_order",
            "coordinates",
            "connections",
            "connections_order",
            "initial_conditions",
            "initial_conditions_order",
        }
    }


def write(
    filename,
    mesh,
    nodal_distance="line",
    material_name=None,
    material_end=None,
    incon=False,
    coord=False,
    eos=None,
    gravity=None,
):
    """Write TOUGH MESH file (and INCON file)."""
    if nodal_distance not in {"line", "orthogonal"}:
        raise ValueError()
    if not (material_name is None or isinstance(material_name, dict)):
        raise TypeError()
    if not (material_end is None or isinstance(material_end, (str, list, tuple))):
        raise TypeError()
    if not isinstance(incon, bool):
        raise TypeError()
    if not (gravity is None or (np.ndim(gravity) == 1 and len(gravity) == 3)):
        raise ValueError()

    # Required variables for blocks ELEME and CONNE
    num_cells = mesh.n_cells
    labels = mesh.labels
    nodes = mesh.centers
    materials = mesh.materials
    volumes = mesh.volumes
    boundary_conditions = (
        mesh.cell_data["boundary_condition"]
        if "boundary_condition" in mesh.cell_data.keys()
        else np.zeros(num_cells, dtype=int)
    )
    points = mesh.points
    connections = mesh.connections
    gravity = gravity if gravity is not None else np.array([0.0, 0.0, -1.0])

    # Define parameters related to faces
    faces = mesh.faces
    face_normals = mesh.face_normals
    face_areas = mesh.face_areas

    # Required variables for block INCON
    primary_variables, porosities, permeabilities, phase_compositions = init_incon(mesh)
    incon = check_incon(
        incon,
        primary_variables,
        porosities,
        permeabilities,
        phase_compositions,
        num_cells,
        eos,
    )

    # Write MESH file
    write_mesh(
        filename,
        num_cells,
        labels,
        nodes,
        materials,
        volumes,
        boundary_conditions,
        points,
        connections,
        gravity,
        faces,
        face_normals,
        face_areas,
        nodal_distance,
        material_name,
        material_end,
        coord,
    )

    # Write INCON file
    if incon:
        head = os.path.split(filename)[0]
        write_incon(
            os.path.join(head, "INCON") if head else "INCON",
            labels,
            primary_variables,
            porosities,
            permeabilities,
            phase_compositions,
            eos,
        )


def check_incon(
    incon,
    primary_variables,
    porosities,
    permeabilities,
    phase_compositions,
    num_cells,
    eos,
):
    """Check INCON inputs and show warnings if necessary."""
    do_incon = incon
    if incon and (primary_variables == -1.0e9).all():
        logging.warning("Initial conditions are not defined. Skipping INCON.")
        do_incon = False

    cond = np.logical_and(
        primary_variables[:, 0] > -1.0e9, primary_variables[:, 0] < 0.0,
    )
    if cond.any():
        logging.warning("Negative pore pressures found in 'INCON'.")

    if porosities is not None:
        if not incon:
            logging.warning("Porosity is only exported if incon is provided.")
        else:
            if not (len(porosities) == num_cells and porosities.ndim == 1):
                raise ValueError("Inconsistent porosity array.")

    if permeabilities is not None:
        if not incon:
            logging.warning(
                "Permeability modifiers are only exported if incon is provided."
            )
        else:
            if permeabilities.ndim not in {1, 2}:
                raise ValueError()
            if not (
                permeabilities.shape == (num_cells, 3)
                if permeabilities.ndim == 2
                else len(permeabilities) == num_cells
            ):
                raise ValueError("Inconsistent permeability modifiers array.")

    if eos == "tmvoc" and phase_compositions is not None:
        if not incon:
            logging.warning("Phase composition is only exported if incon is provided.")
        else:
            if not (
                len(phase_compositions) == num_cells and phase_compositions.ndim == 1
            ):
                raise ValueError("Inconsistent phase composition array.")

    return do_incon


def write_mesh(
    filename,
    num_cells,
    labels,
    nodes,
    materials,
    volumes,
    boundary_conditions,
    points,
    connections,
    gravity,
    faces,
    face_normals,
    face_areas,
    nodal_distance,
    material_name,
    material_end,
    coord,
):
    """Write MESH file."""
    # Check materials
    materials = [
        "{:5}".format(material.strip()) if isinstance(material, str) else material
        for material in materials
    ]
    material_name = material_name if material_name else {}
    material_end = material_end if material_end else []
    material_end = [material_end] if isinstance(material_end, str) else material_end

    with open_file(filename, "w") as f:
        _write_eleme(
            f,
            labels,
            nodes,
            materials,
            volumes,
            boundary_conditions,
            material_name,
            material_end,
        )

        if coord:
            _write_coord(
                f, nodes, materials, material_end,
            )

        _write_conne(
            f,
            labels,
            nodes,
            points,
            connections,
            gravity,
            boundary_conditions,
            faces,
            face_normals,
            face_areas,
            nodal_distance,
        )


def write_incon(
    filename,
    labels,
    primary_variables,
    porosities,
    permeabilities,
    phase_compositions,
    eos,
):
    """Write INCON file."""
    with open_file(filename, "w") as f:
        _write_incon(
            f,
            labels,
            primary_variables,
            porosities,
            permeabilities,
            phase_compositions,
            eos,
        )


@block("ELEME")
def _write_eleme(
    f,
    labels,
    nodes,
    materials,
    volumes,
    boundary_conditions,
    material_name,
    material_end,
):
    """Write ELEME block."""
    from ._helpers import _write_eleme as writer

    # Apply time-independent Dirichlet boundary conditions
    volumes[boundary_conditions.astype(bool)] *= 1.0e50

    # Write ELEME block
    ending = []
    iterables = zip(materials, writer(labels, materials, volumes, nodes, material_name))
    for material, record in iterables:
        if material not in material_end:
            f.write(record)
        else:
            ending.append(record)

    # Append ending cells at the end of the block
    if material_end:
        for record in ending:
            f.write(record)


@block("COORD")
def _write_coord(
    f, nodes, materials, material_end,
):
    """Write COORD block."""
    from ._helpers import _write_coord as writer

    # Write COORD block
    ending = []
    for material, record in zip(materials, writer(nodes)):
        if material not in material_end:
            f.write(record)
        else:
            ending.append(record)

    # Append ending cells at the end of the block
    if material_end:
        for record in ending:
            f.write(record)


@block("CONNE")
def _write_conne(
    f,
    labels,
    nodes,
    points,
    connections,
    gravity,
    boundary_conditions,
    faces,
    face_normals,
    face_areas,
    nodal_distance,
):
    """Write CONNE block."""
    from ._helpers import _write_conne as writer

    # Define unique connection variables
    cell_list = set()
    clabels, centers, int_points, int_normals, areas, bounds = [], [], [], [], [], []
    for i, connection in enumerate(connections):
        if (connection >= 0).any():
            for iface, j in enumerate(connection):
                if j >= 0 and j not in cell_list:
                    # Label
                    clabels.append((labels[i], labels[j]))

                    # Nodal points
                    centers.append([nodes[i], nodes[j]])

                    # Common interface defined by single point and normal vector
                    face = faces[i, iface]
                    int_points.append(points[face[face >= 0][0]])
                    int_normals.append(face_normals[i][iface])

                    # Area of common face
                    areas.append(face_areas[i][iface])

                    # Boundary conditions
                    bounds.append((boundary_conditions[i], boundary_conditions[j]))
        else:
            logging.warning(
                "Element '{}' is not connected to the grid.".format(labels[i])
            )
        cell_list.add(i)

    centers = np.array(centers)
    int_points = np.array(int_points)
    int_normals = np.array(int_normals)
    bounds = np.array(bounds)

    # Calculate remaining variables
    lines = np.diff(centers, axis=1)[:, 0]
    isot = _isot(lines)
    angles = np.dot(lines, gravity) / np.linalg.norm(lines, axis=1)

    if nodal_distance == "line":
        fp = _intersection_line_plane(centers[:, 0], lines, int_points, int_normals)
        d1 = np.where(bounds[:, 0], 1.0e-9, np.linalg.norm(centers[:, 0] - fp, axis=1))
        d2 = np.where(bounds[:, 1], 1.0e-9, np.linalg.norm(centers[:, 1] - fp, axis=1))
    elif nodal_distance == "orthogonal":
        d1 = _distance_point_plane(centers[:, 0], int_points, int_normals, bounds[:, 0])
        d2 = _distance_point_plane(centers[:, 1], int_points, int_normals, bounds[:, 1])

    # Write CONNE block
    for line in writer(clabels, isot, d1, d2, areas, angles):
        f.write(line)


@block("INCON")
def _write_incon(
    f, labels, primary_variables, porosities, permeabilities, phase_compositions, eos
):
    """Write INCON block."""
    from ._helpers import _write_incon as writer

    # Write INCON block
    if permeabilities is not None:
        permeabilities = (
            permeabilities[:, None] if permeabilities.ndim == 1 else permeabilities
        )

    for record in writer(
        labels, primary_variables, porosities, permeabilities, phase_compositions, eos
    ):
        f.write(record)


def init_incon(mesh):
    """Initialize primary variables, porosity and permeability arrays."""
    primary_variables = (
        mesh.cell_data["initial_condition"]
        if "initial_condition" in mesh.cell_data.keys()
        else np.full((mesh.n_cells, 4), -1.0e9)
    )
    porosities = (
        mesh.cell_data["porosity"] if "porosity" in mesh.cell_data.keys() else None
    )
    permeabilities = (
        mesh.cell_data["permeability"]
        if "permeability" in mesh.cell_data.keys()
        else None
    )
    phase_compositions = (
        mesh.cell_data["phase_composition"]
        if "phase_composition" in mesh.cell_data.keys()
        else None
    )

    return primary_variables, porosities, permeabilities, phase_compositions


def _intersection_line_plane(center, lines, int_points, int_normals):
    """
    Calculate the intersection point between a line and a plane.

    Calculate intersection between a line defined by a point and a direction vector and
    a plane defined by one point and a normal vector.

    """
    tmp = _dot(int_points - center, int_normals) / _dot(lines, int_normals)
    return center + lines * tmp[:, None]


def _distance_point_plane(center, int_points, int_normals, mask):
    """
    Calculate orthogonal distance.

    Calculate orthogonal distance of a point to a plane defined by one point and a
    normal vector.

    """
    return np.where(mask, 1.0e-9, np.abs(_dot(center - int_points, int_normals)))


def _isot(lines):
    """
    Determine direction of anisotropy.

    Given the direction of the line connecting the cell centers, calculate the
    direction of anisotropy.

    Note
    ----
    It always returns 1 if the connection line is not colinear with X, Y or Z.

    """
    mask = lines != 0.0
    return np.where(mask.sum(axis=1) == 1, mask.argmax(axis=1) + 1, 1)


def _dot(A, B):
    """Calculate dot product when arrays A and B have the same shape."""
    return (A * B).sum(axis=1)



import helpers
import numpy as np
import pytest

import toughio

write_read = lambda x, **kwargs: helpers.write_read(
    "INFILE", x, toughio.write_input, toughio.read_input, **kwargs
)

write_read_tough = lambda x: write_read(
    x, writer_kws={"file_format": "tough"}, reader_kws={"file_format": "tough"},
)

write_read_json = lambda x: write_read(
    x, writer_kws={"file_format": "json"}, reader_kws={"file_format": "json"},
)


@pytest.mark.parametrize(
    "write_read, single",
    [
        (write_read_tough, True),
        (write_read_tough, False),
        (write_read_json, True),
        (write_read_json, False),
    ],
)
def test_title(write_read, single):
    parameters_ref = {
        "title": (
            helpers.random_string(80)
            if single
            else [helpers.random_string(80) for _ in range(np.random.randint(5) + 2)]
        ),
    }
    parameters = write_read(parameters_ref)

    assert parameters_ref["title"] == parameters["title"]


@pytest.mark.parametrize("write_read", [write_read_tough, write_read_json])
def test_rocks(write_read):
    keys = [
        "density",
        "porosity",
        "permeability",
        "conductivity",
        "specific_heat",
        "compressibility",
        "expansivity",
        "conductivity_dry",
        "tortuosity",
        "klinkenberg_parameter",
        "distribution_coefficient_3",
        "distribution_coefficient_4",
    ]
    parameters_ref = {
        "rocks": {
            helpers.random_string(5): {key: np.random.rand() for key in keys[:5]},
            helpers.random_string(5): {
                key: np.random.rand() if key != "permeability" else np.random.rand(3)
                for key in keys[:5]
            },
            helpers.random_string(5): {key: np.random.rand() for key in keys},
            helpers.random_string(5): {key: np.random.rand() for key in keys},
            helpers.random_string(5): {key: np.random.rand() for key in keys},
            helpers.random_string(5): {key: np.random.rand() for key in keys},
        }
    }
    names = list(parameters_ref["rocks"].keys())
    parameters_ref["rocks"][names[-1]].update(
        {
            "relative_permeability": {
                "id": np.random.randint(10),
                "parameters": np.random.rand(np.random.randint(7) + 1),
            },
        }
    )
    parameters_ref["rocks"][names[-2]].update(
        {
            "capillarity": {
                "id": np.random.randint(10),
                "parameters": np.random.rand(np.random.randint(7) + 1),
            },
        }
    )
    parameters_ref["rocks"][names[-3]].update(
        {
            "relative_permeability": {
                "id": np.random.randint(10),
                "parameters": np.random.rand(np.random.randint(7) + 1),
            },
            "capillarity": {
                "id": np.random.randint(10),
                "parameters": np.random.rand(np.random.randint(7) + 1),
            },
        }
    )
    parameters = write_read(parameters_ref)

    assert sorted(parameters_ref["rocks"].keys()) == sorted(parameters["rocks"].keys())

    for k, v in parameters_ref["rocks"].items():
        for kk, vv in v.items():
            if not isinstance(vv, dict):
                assert np.allclose(vv, parameters["rocks"][k][kk], atol=1.0e-5)
            else:
                helpers.allclose_dict(vv, parameters["rocks"][k][kk], atol=1.0e-4)


@pytest.mark.parametrize(
    "write_read, rpcap",
    [
        (write_read_tough, "rp"),
        (write_read_tough, "cap"),
        (write_read_tough, "both"),
        (write_read_json, "rp"),
        (write_read_json, "cap"),
        (write_read_json, "both"),
    ],
)
def test_rpcap(write_read, rpcap):
    parameters_ref = {"default": {}}
    if rpcap in {"rp", "both"}:
        parameters_ref["default"]["relative_permeability"] = {
            "id": np.random.randint(10),
            "parameters": np.random.rand(np.random.randint(7) + 1),
        }
    if rpcap in {"cap", "both"}:
        parameters_ref["default"]["capillarity"] = {
            "id": np.random.randint(10),
            "parameters": np.random.rand(np.random.randint(7) + 1),
        }
    parameters = write_read(parameters_ref)

    for k, v in parameters_ref["default"].items():
        helpers.allclose_dict(v, parameters["default"][k], atol=1.0e-4)


@pytest.mark.parametrize("write_read", [write_read_tough, write_read_json])
def test_flac(write_read):
    parameters_ref = {
        "flac": {
            "creep": bool(np.random.randint(2)),
            "porosity_model": np.random.randint(10),
            "version": np.random.randint(10),
        },
        "rocks": {
            helpers.random_string(5): {
                "permeability_model": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(np.random.randint(7) + 1),
                },
                "equivalent_pore_pressure": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(np.random.randint(7) + 1),
                },
            }
            for _ in np.random.rand(10) + 1
        },
    }
    parameters = write_read(parameters_ref)

    helpers.allclose_dict(parameters_ref["flac"], parameters["flac"])
    for k, v in parameters_ref["rocks"].items():
        for kk, vv in v.items():
            helpers.allclose_dict(vv, parameters["rocks"][k][kk], atol=1.0e-4)


@pytest.mark.parametrize("write_read", [write_read_tough, write_read_json])
def test_chemp(write_read):
    parameters_ref = {
        "chemical_properties": {
            helpers.random_string(20): {
                "temperature_crit": np.random.rand(),
                "pressure_crit": np.random.rand(),
                "compressibility_crit": np.random.rand(),
                "pitzer_factor": np.random.rand(),
                "dipole_moment": np.random.rand(),
                "boiling_point": np.random.rand(),
                "vapor_pressure_a": np.random.rand(),
                "vapor_pressure_b": np.random.rand(),
                "vapor_pressure_c": np.random.rand(),
                "vapor_pressure_d": np.random.rand(),
                "molecular_weight": np.random.rand(),
                "heat_capacity_a": np.random.rand(),
                "heat_capacity_b": np.random.rand(),
                "heat_capacity_c": np.random.rand(),
                "heat_capacity_d": np.random.rand(),
                "napl_density_ref": np.random.rand(),
                "napl_temperature_ref": np.random.rand(),
                "gas_diffusivity_ref": np.random.rand(),
                "gas_temperature_ref": np.random.rand(),
                "exponent": np.random.rand(),
                "napl_viscosity_a": np.random.rand(),
                "napl_viscosity_b": np.random.rand(),
                "napl_viscosity_c": np.random.rand(),
                "napl_viscosity_d": np.random.rand(),
                "volume_crit": np.random.rand(),
                "solubility_a": np.random.rand(),
                "solubility_b": np.random.rand(),
                "solubility_c": np.random.rand(),
                "solubility_d": np.random.rand(),
                "oc_coeff": np.random.rand(),
                "oc_fraction": np.random.rand(),
                "oc_decay": np.random.rand(),
            }
            for _ in np.random.rand(10) + 1
        }
    }
    parameters = write_read(parameters_ref)

    assert len(parameters["chemical_properties"]) == len(
        parameters_ref["chemical_properties"]
    )
    for k, v in parameters_ref["chemical_properties"].items():
        for kk, vv in v.items():
            assert np.allclose(
                vv, parameters["chemical_properties"][k][kk], atol=1.0e-4
            )


@pytest.mark.parametrize("write_read", [write_read_tough, write_read_json])
def test_ncgas(write_read):
    parameters_ref = {
        "non_condensible_gas": [
            helpers.random_string(10) for _ in np.random.rand(10) + 1
        ]
    }
    parameters = write_read(parameters_ref)

    assert len(parameters["non_condensible_gas"]) == len(
        parameters_ref["non_condensible_gas"]
    )
    for v1, v2 in zip(
        parameters["non_condensible_gas"], parameters_ref["non_condensible_gas"]
    ):
        assert v1 == v2


@pytest.mark.parametrize(
    "write_read, isothermal", [(write_read_tough, True), (write_read_tough, False)],
)
def test_multi(write_read, isothermal):
    import random

    from toughio._io.input.tough._common import eos

    parameters_ref = {
        "eos": random.choice(
            [k for k in eos.keys() if k not in {"eos7", "eos8", "eos9", "tmvoc"}]
        ),
        "isothermal": isothermal,
    }
    parameters = write_read(parameters_ref)

    multi = [
        parameters["n_component"],
        parameters["n_component"] + 1,
        parameters["n_phase"],
        6,
    ]
    multi_ref = eos[parameters_ref["eos"]]
    assert multi_ref == multi
    assert parameters_ref["isothermal"] == parameters["isothermal"]


@pytest.mark.parametrize("write_read", [write_read_tough, write_read_json])
def test_solvr(write_read):
    parameters_ref = {
        "solver": {
            "method": np.random.randint(10),
            "z_precond": helpers.random_string(2),
            "o_precond": helpers.random_string(2),
            "rel_iter_max": np.random.rand(),
            "eps": np.random.rand(),
        },
    }
    parameters = write_read(parameters_ref)

    assert parameters_ref["solver"]["method"] == parameters["solver"]["method"]
    assert parameters_ref["solver"]["z_precond"] == parameters["solver"]["z_precond"]
    assert parameters_ref["solver"]["o_precond"] == parameters["solver"]["o_precond"]
    assert np.allclose(
        parameters_ref["solver"]["rel_iter_max"],
        parameters["solver"]["rel_iter_max"],
        atol=1.0e-5,
    )
    assert np.allclose(
        parameters_ref["solver"]["eps"], parameters["solver"]["eps"], atol=1.0e-5
    )


@pytest.mark.parametrize(
    "write_read, t_steps, num_pvars",
    [
        (write_read_tough, np.random.rand(), 4),
        (write_read_tough, np.random.rand(np.random.randint(100) + 1), 4),
        (write_read_tough, np.random.rand(np.random.randint(100) + 1), 6),
        (write_read_json, np.random.rand(), 4),
        (write_read_json, np.random.rand(np.random.randint(100) + 1), 4),
        (write_read_json, np.random.rand(np.random.randint(100) + 1), 6),
    ],
)
def test_param(write_read, t_steps, num_pvars):
    parameters_ref = {
        "options": {
            "n_iteration": np.random.randint(10),
            "n_cycle": np.random.randint(10),
            "n_second": np.random.randint(10),
            "n_cycle_print": np.random.randint(10),
            "verbosity": np.random.randint(10),
            "temperature_dependence_gas": np.random.rand(),
            "effective_strength_vapor": np.random.rand(),
            "t_ini": np.random.rand(),
            "t_max": np.random.rand(),
            "t_steps": t_steps,
            "t_step_max": np.random.rand(),
            "t_reduce_factor": np.random.rand(),
            "gravity": np.random.rand(),
            "mesh_scale_factor": np.random.rand(),
            "eps1": np.random.rand(),
            "eps2": np.random.rand(),
            "w_upstream": np.random.rand(),
            "w_newton": np.random.rand(),
            "derivative_factor": np.random.rand(),
        },
        "extra_options": {
            k + 1: v for k, v in enumerate(np.random.randint(10, size=24))
        },
        "default": {"initial_condition": np.random.rand(num_pvars)},
    }
    parameters = write_read(parameters_ref)

    helpers.allclose_dict(parameters_ref["options"], parameters["options"], atol=1.0e-5)
    helpers.allclose_dict(parameters_ref["extra_options"], parameters["extra_options"])
    if "initial_condition" in parameters["default"].keys():
        assert np.allclose(
            parameters_ref["default"]["initial_condition"],
            parameters["default"]["initial_condition"],
            atol=1.0e-5,
        )
    else:
        assert not len(parameters_ref["default"]["initial_condition"])


@pytest.mark.parametrize(
    "write_read, num_floats",
    [
        (write_read_tough, None),
        (write_read_tough, 8),
        (write_read_json, None),
        (write_read_json, 8),
    ],
)
def test_selec(write_read, num_floats):
    parameters_ref = {
        "selections": {
            "integers": {
                k + 1: v for k, v in enumerate(np.random.randint(100, size=16))
            },
            "floats": (
                np.random.rand(num_floats)
                if num_floats is not None and num_floats <= 8
                else np.random.rand(
                    np.random.randint(100) + 1, np.random.randint(8) + 1
                )
            ),
        },
    }
    parameters_ref["selections"]["integers"][1] = (
        len(parameters_ref["selections"]["floats"])
        if np.ndim(parameters_ref["selections"]["floats"]) == 2
        else 1
    )
    parameters = write_read(parameters_ref)

    helpers.allclose_dict(
        parameters_ref["selections"]["integers"], parameters["selections"]["integers"]
    )
    if "floats" in parameters["selections"].keys():
        assert np.allclose(
            parameters_ref["selections"]["floats"],
            parameters["selections"]["floats"],
            atol=1.0e-4,
        )
    else:
        assert parameters_ref["selections"]["integers"][1] == 0


@pytest.mark.parametrize(
    "write_read, num_pvars, num_items",
    [
        (write_read_tough, 4, None),
        (write_read_tough, 6, None),
        (write_read_tough, 4, 1),
        (write_read_tough, 6, 1),
        (write_read_json, 4, None),
        (write_read_json, 6, None),
    ],
)
def test_indom(write_read, num_pvars, num_items):
    num_items = num_items if num_items else np.random.randint(10) + 1
    parameters_ref = {
        "rocks": {
            helpers.random_string(5): {"initial_condition": np.random.rand(num_pvars),}
            for _ in range(num_items)
        },
    }
    parameters = write_read(parameters_ref)

    for k, v in parameters_ref["rocks"].items():
        assert np.allclose(
            v["initial_condition"],
            parameters["rocks"][k]["initial_condition"],
            atol=1.0e-4,
        )


@pytest.mark.parametrize("write_read", [write_read_tough, write_read_json])
def test_momop(write_read):
    parameters_ref = {
        "more_options": {
            k + 1: v for k, v in enumerate(np.random.randint(10, size=40))
        },
    }
    parameters = write_read(parameters_ref)

    helpers.allclose_dict(parameters_ref["more_options"], parameters["more_options"])


@pytest.mark.parametrize(
    "write_read, times",
    [
        (write_read_tough, np.random.rand(np.random.randint(100) + 1)),
        (write_read_json, np.random.rand(np.random.randint(100) + 1)),
    ],
)
def test_times(write_read, times):
    parameters_ref = {"times": times}
    parameters = write_read(parameters_ref)

    assert np.allclose(parameters_ref["times"], parameters["times"], atol=1.0e-5)


@pytest.mark.parametrize(
    "write_read, oft, n",
    [
        (write_read_tough, "element_history", 5),
        (write_read_tough, "connection_history", 10),
        (write_read_tough, "generator_history", 5),
        (write_read_json, "element_history", 5),
        (write_read_json, "connection_history", 10),
        (write_read_json, "generator_history", 5),
    ],
)
def test_oft(write_read, oft, n):
    parameters_ref = {
        oft: [helpers.random_string(n) for _ in range(np.random.randint(10) + 1)]
    }
    parameters = write_read(parameters_ref)

    assert parameters_ref[oft] == parameters[oft]


@pytest.mark.parametrize(
    "write_read, specific_enthalpy, label_length",
    [
        (write_read_tough, True, 5),
        (write_read_json, True, 5),
        (write_read_tough, True, 6),
        (write_read_json, True, 6),
        (write_read_tough, False, 5),
        (write_read_json, False, 5),
        (write_read_tough, False, 6),
        (write_read_json, False, 6),
    ],
)
def test_gener(write_read, specific_enthalpy, label_length):
    n_rnd = np.random.randint(100) + 2
    parameters_ref = {
        "generators": [
            {
                "label": helpers.random_label(label_length),
                "name": helpers.random_string(5),
                "nseq": np.random.randint(10),
                "nadd": np.random.randint(10),
                "nads": np.random.randint(10),
                "type": helpers.random_string(4),
                "rates": np.random.rand(),
                "specific_enthalpy": np.random.rand(),
                "layer_thickness": np.random.rand(),
            },
            {
                "label": helpers.random_label(label_length),
                "nseq": np.random.randint(10),
                "nadd": np.random.randint(10),
                "nads": np.random.randint(10),
                "type": helpers.random_string(4),
                "times": np.random.rand(n_rnd),
                "rates": np.random.rand(n_rnd),
                "specific_enthalpy": np.random.rand(n_rnd),
                "layer_thickness": np.random.rand(),
            },
        ],
    }
    parameters = write_read(parameters_ref)

    assert len(parameters_ref["generators"]) == len(parameters["generators"])

    for generator_ref, generator in zip(
        parameters_ref["generators"], parameters["generators"]
    ):
        for k, v in generator_ref.items():
            if k in {"label", "name", "type"}:
                assert v == generator[k]

            else:
                assert np.allclose(v, generator[k], atol=1.0e-4)


@pytest.mark.parametrize("write_read", [write_read_tough, write_read_json])
def test_diffu(write_read):
    n_phase = np.random.randint(8) + 1
    parameters_ref = {
        "n_phase": n_phase,
        "diffusion": np.random.rand(np.random.randint(5) + 1, n_phase),
    }
    parameters = write_read(parameters_ref)

    assert np.allclose(
        parameters_ref["diffusion"], parameters["diffusion"], atol=1.0e-4
    )


@pytest.mark.parametrize(
    "write_read, fmt",
    [
        (write_read_tough, None),
        (write_read_tough, helpers.random_string(20)),
        (write_read_json, None),
        (write_read_json, helpers.random_string(20)),
    ],
)
def test_outpu(write_read, fmt):
    parameters_ref = {
        "output": {
            "format": fmt,
            "variables": [
                {"name": helpers.random_string(20)},
                {"name": helpers.random_string(20), "options": None},
                {"name": helpers.random_string(20), "options": np.random.randint(10)},
                {
                    "name": helpers.random_string(20),
                    "options": np.random.randint(10, size=1),
                },
                {
                    "name": helpers.random_string(20),
                    "options": np.random.randint(10, size=2),
                },
            ],
        },
    }
    parameters = write_read(parameters_ref)

    for variable_ref, variable in zip(
        parameters_ref["output"]["variables"], parameters["output"]["variables"]
    ):
        assert variable_ref["name"] == variable["name"]

        if "options" in variable_ref and variable_ref["options"] is not None:
            assert np.allclose(variable_ref["options"], variable["options"])


@pytest.mark.parametrize(
    "write_read, label_length, coord",
    [
        (write_read_tough, 5, False),
        (write_read_json, 5, False),
        (write_read_tough, 6, False),
        (write_read_json, 6, False),
        (write_read_tough, 5, True),
        (write_read_json, 5, True),
        (write_read_tough, 6, True),
        (write_read_json, 6, True),
    ],
)
def test_eleme(write_read, label_length, coord):
    labels = [
        helpers.random_label(label_length) for _ in range(np.random.randint(10) + 1)
    ]
    keys = [
        "nseq",
        "nadd",
        "material",
        "volume",
        "heat_exchange_area",
        "permeability_modifier",
        "center",
    ]
    parameters_ref = {
        "elements": {
            label: {
                key: (
                    np.random.randint(10)
                    if key in {"nseq", "nadd"}
                    else helpers.random_string(5)
                    if key == "material"
                    else np.random.rand(3)
                    if key == "center"
                    else np.random.rand()
                )
                for key in keys
            }
            for label in labels
        },
        "coordinates": coord,
    }
    parameters = write_read(parameters_ref)

    assert sorted(parameters_ref["elements"].keys()) == sorted(
        parameters["elements"].keys()
    )

    for k, v in parameters_ref["elements"].items():
        for kk, vv in v.items():
            if not isinstance(vv, str):
                assert np.allclose(vv, parameters["elements"][k][kk], atol=1.0e-3)
            else:
                assert vv == parameters["elements"][k][kk]

    assert parameters_ref["coordinates"] == parameters["coordinates"]


@pytest.mark.parametrize(
    "write_read, label_length",
    [
        (write_read_tough, 5),
        (write_read_json, 5),
        (write_read_tough, 6),
        (write_read_json, 6),
    ],
)
def test_conne(write_read, label_length):
    labels = [
        "".join(helpers.random_label(label_length) for _ in range(2))
        for _ in range(np.random.randint(10) + 1)
    ]
    keys = [
        "nseq",
        "nadd",
        "permeability_direction",
        "nodal_distances",
        "interface_area",
        "gravity_cosine_angle",
        "radiant_emittance_factor",
    ]
    parameters_ref = {
        "connections": {
            label: {
                key: (
                    np.random.randint(10)
                    if key == "nseq"
                    else np.random.randint(10, size=2)
                    if key == "nadd"
                    else np.random.randint(1, 4)
                    if key == "permeability_direction"
                    else np.random.rand(2)
                    if key == "nodal_distances"
                    else np.random.rand()
                )
                for key in keys
            }
            for label in labels
        }
    }
    parameters = write_read(parameters_ref)

    assert sorted(parameters_ref["connections"].keys()) == sorted(
        parameters["connections"].keys()
    )

    for k, v in parameters_ref["connections"].items():
        for kk, vv in v.items():
            assert np.allclose(vv, parameters["connections"][k][kk], atol=1.0e-4)


@pytest.mark.parametrize(
    "write_read, label_length, num_pvars, num_items",
    [
        (write_read_tough, 5, 4, None),
        (write_read_tough, 5, 6, None),
        (write_read_tough, 6, 4, None),
        (write_read_tough, 5, 4, 1),
        (write_read_tough, 5, 6, 1),
        (write_read_json, 5, 4, None),
        (write_read_json, 5, 6, None),
        (write_read_json, 6, 4, None),
    ],
)
def test_incon(write_read, label_length, num_pvars, num_items):
    num_items = num_items if num_items else np.random.randint(10) + 1
    labels = [helpers.random_label(label_length) for _ in range(num_items)]
    keys = [
        "porosity",
        "userx",
        "values",
    ]
    parameters_ref = {
        "initial_conditions": {
            label: {
                key: (
                    np.random.rand()
                    if key == "porosity"
                    else np.random.rand(np.random.randint(5) + 1)
                    if key == "userx"
                    else np.random.rand(num_pvars)
                )
                for key in keys
            }
            for label in labels
        }
    }
    parameters = write_read(parameters_ref)

    assert sorted(parameters_ref["initial_conditions"].keys()) == sorted(
        parameters["initial_conditions"].keys()
    )

    for k, v in parameters_ref["initial_conditions"].items():
        for kk, vv in v.items():
            assert np.allclose(vv, parameters["initial_conditions"][k][kk], atol=1.0e-3)


def test_meshm_xyz():
    parameters_ref = {
        "meshmaker": {
            "type": "xyz",
            "parameters": [
                {
                    "type": "nx",
                    "n_increment": np.random.randint(100) + 1,
                    "sizes": np.random.rand(),
                },
                {"type": "ny", "sizes": np.random.rand(np.random.randint(100) + 1),},
                {"type": "nz", "sizes": np.random.rand(np.random.randint(100) + 1),},
                {"type": "nx", "sizes": np.random.rand(np.random.randint(100) + 1),},
            ],
            "angle": np.random.rand(),
        }
    }
    parameters = write_read(parameters_ref)

    assert parameters_ref["meshmaker"]["type"] == parameters["meshmaker"]["type"]
    assert np.allclose(
        parameters_ref["meshmaker"]["angle"],
        parameters["meshmaker"]["angle"],
        atol=1.0e-4,
    )
    assert len(parameters_ref["meshmaker"]["parameters"]) == len(
        parameters_ref["meshmaker"]["parameters"]
    )

    for parameter_ref, parameter in zip(
        parameters_ref["meshmaker"]["parameters"], parameters["meshmaker"]["parameters"]
    ):
        for k, v in parameter_ref.items():
            if isinstance(v, str):
                assert v == parameter[k]

            else:
                assert np.allclose(v, parameter[k], atol=1.0e-4)


@pytest.mark.parametrize("layer", [True, False])
def test_meshm_rz2d(layer):
    parameters_ref = {
        "meshmaker": {
            "type": "rz2dl" if layer else "rz2d",
            "parameters": [
                {"type": "radii", "radii": np.random.rand(np.random.randint(100) + 1),},
                {
                    "type": "equid",
                    "n_increment": np.random.randint(100) + 1,
                    "size": np.random.rand(),
                },
                {
                    "type": "logar",
                    "n_increment": np.random.randint(100) + 1,
                    "radius": np.random.rand(),
                    "radius_ref": np.random.rand(),
                },
                {
                    "type": "layer",
                    "thicknesses": np.random.rand(np.random.randint(100) + 1),
                },
            ],
        }
    }
    parameters = write_read(parameters_ref)

    assert parameters_ref["meshmaker"]["type"] == parameters["meshmaker"]["type"]
    assert len(parameters_ref["meshmaker"]["parameters"]) == len(
        parameters_ref["meshmaker"]["parameters"]
    )

    for parameter_ref, parameter in zip(
        parameters_ref["meshmaker"]["parameters"], parameters["meshmaker"]["parameters"]
    ):
        for k, v in parameter_ref.items():
            if isinstance(v, str):
                assert v == parameter[k]

            else:
                assert np.allclose(v, parameter[k], atol=1.0e-4)


def test_tmvoc():
    parameters_ref = {
        "eos": "tmvoc",
        "n_component": 1,
        "n_phase": 1,
        "default": {"phase_composition": np.random.randint(10),},
        "rocks": {
            helpers.random_string(5): {
                "initial_condition": np.random.rand(4),
                "phase_composition": np.random.randint(10),
            }
            for _ in range(np.random.randint(10) + 1)
        },
        "initial_conditions": {
            helpers.random_string(5): {
                "values": np.random.rand(4),
                "phase_composition": np.random.randint(10),
            }
            for _ in range(np.random.randint(10) + 1)
        },
    }
    parameters = write_read(
        parameters_ref, writer_kws={"eos": "tmvoc"}, reader_kws={"eos": "tmvoc"},
    )

    helpers.allclose_dict(parameters_ref["default"], parameters["default"])

    assert sorted(parameters_ref["rocks"].keys()) == sorted(parameters["rocks"].keys())
    for k, v in parameters_ref["rocks"].items():
        helpers.allclose_dict(v, parameters["rocks"][k])

    assert sorted(parameters_ref["initial_conditions"].keys()) == sorted(
        parameters["initial_conditions"].keys()
    )
    for k, v in parameters_ref["initial_conditions"].items():
        helpers.allclose_dict(v, parameters["initial_conditions"][k])


@pytest.mark.parametrize(
    "write_read, flag, enable",
    [
        (write_read_tough, "start", True),
        (write_read_tough, "start", False),
        (write_read_tough, "nover", True),
        (write_read_tough, "nover", False),
        (write_read_json, "start", True),
        (write_read_json, "start", False),
        (write_read_json, "nover", True),
        (write_read_json, "nover", False),
    ],
)
def test_flag(write_read, flag, enable):
    parameters_ref = {flag: enable}
    parameters = write_read(parameters_ref)

    if flag in parameters.keys():
        assert parameters_ref[flag] == parameters[flag]
    else:
        assert not enable



from . import csv, petrasim, save, tecplot, tough
from ._common import Output
from ._helpers import read, read_history, register, write

__all__ = [
    "Output",
    "register",
    "read",
    "write",
    "read_history",
]



from __future__ import division, with_statement

import logging
from copy import deepcopy

import numpy as np

from ...._common import block_to_format, open_file, str2format
from ._common import default
from ._helpers import block, check_parameters, dtypes, prune_nones_list, write_record

__all__ = [
    "write",
]


def write(filename, parameters, block=None, ignore_blocks=None, eos=None):
    """
    Write TOUGH input file.

    Parameters
    ----------
    filename : str
        Output file name.
    parameters : dict
        Parameters to export.
    block : str {'all', 'gener', 'mesh', 'incon'} or None, optional, default None
        Blocks to be written:
         - 'all': write all blocks,
         - 'gener': only write block GENER,
         - 'mesh': only write blocks ELEME, COORD and CONNE,
         - 'incon': only write block INCON,
         - None: write all blocks except blocks defined in `ignore_blocks`.
    ignore_blocks : list of str or None, optional, default None
        Blocks to ignore. Only if `block` is None.
    eos : str or None, optional, default None
        Equation of State. If `eos` is defined in `parameters`, this option will be ignored.

    """
    # Deprecation error
    if "generators" in parameters:
        if isinstance(parameters["generators"], dict):
            raise ValueError("'generators' must be a list of dicts since v1.7.0.")

    if "output" in parameters:
        if "variables" in parameters["output"] and isinstance(
            parameters["output"]["variables"], dict
        ):
            raise ValueError("'variables' must be a list of dicts since v1.7.0.")

    buffer = write_buffer(parameters, block, ignore_blocks, eos)
    with open_file(filename, "w") as f:
        for record in buffer:
            f.write(record)


@check_parameters(dtypes["PARAMETERS"])
def write_buffer(params, block, ignore_blocks=None, eos_=None):
    """Write TOUGH input file as a list of 80-character long record strings."""
    from ._common import Parameters
    from ._common import blocks as blocks_
    from ._common import default, eos

    # Block filters
    if block is not None:
        if block == "all":
            blocks = blocks_.copy()

        elif block == "gener":
            blocks = {"GENER"}

        elif block == "mesh":
            blocks = {"ELEME", "COORD", "CONNE"}

        elif block == "incon":
            blocks = {"INCON"}

        else:
            raise ValueError()

    else:
        blocks = blocks_.copy()

        if ignore_blocks is not None:
            if not isinstance(ignore_blocks, (list, tuple, np.ndarray)):
                raise TypeError()

            ignore_blocks = set(ignore_blocks)
            for ignore_block in ignore_blocks:
                if ignore_block not in blocks_:
                    raise ValueError("unknown block '{}'.".format(ignore_block))

                else:
                    blocks.remove(ignore_block)

    # Some preprocessing
    parameters = deepcopy(Parameters)
    parameters.update(deepcopy(params))

    parameters["title"] = (
        [parameters["title"]]
        if isinstance(parameters["title"], str)
        else parameters["title"]
    )

    for k, v in default.items():
        if k not in parameters["default"].keys():
            parameters["default"][k] = v

    for rock in parameters["rocks"].keys():
        for k, v in parameters["default"].items():
            cond1 = k not in parameters["rocks"][rock].keys()
            cond2 = k not in {
                "initial_condition",
                "relative_permeability",
                "capillarity",
                "phase_composition",
            }
            if cond1 and cond2:
                parameters["rocks"][rock][k] = v

    tmp = parameters["default"]["permeability"]
    if np.ndim(tmp) == 0:
        tmp = 3 * [tmp]

    for v in parameters["rocks"].values():
        if v["permeability"] is None:
            v["permeability"] = tmp

        elif np.ndim(v["permeability"]):
            if len(v["permeability"]) != 3:
                raise ValueError()

            for i, k in enumerate(tmp):
                if v["permeability"][i] is None:
                    v["permeability"][i] = k

    # Make sure that some keys are integers and not strings
    if "more_options" in parameters:
        parameters["more_options"] = {
            int(k): v for k, v in parameters["more_options"].items()
        }
    if "extra_options" in parameters:
        parameters["extra_options"] = {
            int(k): v for k, v in parameters["extra_options"].items()
        }
    if "selections" in parameters and "integers" in parameters["selections"]:
        parameters["selections"]["integers"] = {
            int(k): v for k, v in parameters["selections"]["integers"].items()
        }

    # Check that EOS is defined (for block MULTI)
    if parameters["eos"] and parameters["eos"] not in eos.keys():
        raise ValueError(
            "EOS '{}' is unknown or not supported.".format(parameters["eos"])
        )

    if parameters["eos"]:
        eos_ = parameters["eos"]

    else:
        parameters["eos"] = eos_

    # Set some flags
    cond1 = (
        "relative_permeability" in parameters["default"].keys()
        and parameters["default"]["relative_permeability"]["id"] is not None
    )
    cond2 = (
        "capillarity" in parameters["default"].keys()
        and parameters["default"]["capillarity"]["id"] is not None
    )
    rpcap = cond1 or cond2

    param = False
    if prune_nones_list(parameters["default"]["initial_condition"]):
        param = True
    if eos_ == "tmvoc" and parameters["default"]["phase_composition"] is not None:
        param = True

    for option in parameters["options"].values():
        if param:
            break

        if (isinstance(option, (list, tuple, np.ndarray)) and len(option)) or option:
            param = True

    if parameters["extra_options"]:
        param = True

    indom = False
    for rock in parameters["rocks"].values():
        if "initial_condition" in rock.keys():
            if any(x is not None for x in rock["initial_condition"][:4]):
                indom = True
                break

    multi = (
        parameters["eos"]
        or parameters["n_component"]
        or parameters["n_phase"]
        or parameters["n_component_incon"]
    )

    output = False
    for key in ["format", "variables"]:
        if key in parameters["output"] and parameters["output"][key]:
            output = True
            break

    # Check that start is True if indom is True
    if indom and not parameters["start"]:
        logging.warning("Option 'START' is needed to use 'INDOM' conditions.")

    # Define input file contents
    out = []
    if "TITLE" in blocks:
        out += ["{:80}\n".format(title) for title in parameters["title"]]

    if "ROCKS" in blocks and parameters["rocks"]:
        out += _write_rocks(parameters)

    if "RPCAP" in blocks and rpcap:
        out += _write_rpcap(parameters)

    if "FLAC" in blocks and parameters["flac"]:
        out += _write_flac(parameters)

    if "CHEMP" in blocks and parameters["chemical_properties"]:
        out += _write_chemp(parameters)

    if "NCGAS" in blocks and len(parameters["non_condensible_gas"]):
        out += _write_ncgas(parameters)

    if "MULTI" in blocks and multi:
        out += _write_multi(parameters)

    if "SOLVR" in blocks and parameters["solver"]:
        out += _write_solvr(parameters)

    if "START" in blocks and parameters["start"]:
        out += _write_start()

    if "PARAM" in blocks and param:
        out += _write_param(parameters, eos_)

    if "SELEC" in blocks and parameters["selections"]:
        out += _write_selec(parameters)

    if "INDOM" in blocks and indom:
        out += _write_indom(parameters, eos_)

    if "MOMOP" in blocks and parameters["more_options"]:
        out += _write_momop(parameters)

    if "TIMES" in blocks and len(parameters["times"]):
        out += _write_times(parameters)

    if "FOFT" in blocks and len(parameters["element_history"]):
        out += _write_foft(parameters)

    if "COFT" in blocks and len(parameters["connection_history"]):
        out += _write_coft(parameters)

    if "GOFT" in blocks and len(parameters["generator_history"]):
        out += _write_goft(parameters)

    if "GENER" in blocks and parameters["generators"]:
        out += _write_gener(parameters)

    if "DIFFU" in blocks and len(parameters["diffusion"]):
        out += _write_diffu(parameters)

    if "OUTPU" in blocks and output:
        out += _write_outpu(parameters)

    if "ELEME" in blocks and parameters["elements"]:
        out += _write_eleme(parameters)

    if "COORD" in blocks and parameters["coordinates"]:
        out += _write_coord(parameters)

    if "CONNE" in blocks and parameters["connections"]:
        out += _write_conne(parameters)

    if "INCON" in blocks and parameters["initial_conditions"]:
        out += _write_incon(parameters, eos_)

    if "MESHM" in blocks and parameters["meshmaker"]:
        out += _write_meshm(parameters)

    if "NOVER" in blocks and parameters["nover"]:
        out += _write_nover()

    if "ENDCY" in blocks:
        out += _write_endcy()

    return out


@check_parameters(dtypes["ROCKS"], keys="default")
@check_parameters(dtypes["ROCKS"], keys="rocks", is_list=True)
@check_parameters(
    dtypes["MODEL"], keys=("rocks", "relative_permeability"), is_list=True
)
@check_parameters(dtypes["MODEL"], keys=("rocks", "capillarity"), is_list=True)
@block("ROCKS", multi=True)
def _write_rocks(parameters):
    """Write ROCKS block data."""
    # Reorder rocks
    if parameters["rocks_order"] is not None:
        order = parameters["rocks_order"]
        for rock in parameters["rocks"].keys():
            if rock not in order:
                order.append(rock)
    else:
        order = parameters["rocks"].keys()

    # Formats
    fmt = block_to_format["ROCKS"]
    fmt1 = str2format(fmt[1])
    fmt2 = str2format(fmt[2])

    fmt = block_to_format["RPCAP"]
    fmt3 = str2format(fmt)

    out = []
    for k in order:
        # Load data
        data = parameters["rocks"][k]

        # Number of additional lines to write per rock
        cond = any(
            data[k] is not None
            for k in [
                "compressibility",
                "expansivity",
                "conductivity_dry",
                "tortuosity",
                "klinkenberg_parameter",
                "distribution_coefficient_3",
                "distribution_coefficient_4",
            ]
        )
        nad = (
            2
            if "relative_permeability" in data.keys() or "capillarity" in data.keys()
            else int(cond)
        )

        # Permeability
        per = data["permeability"]
        per = [per] * 3 if not np.ndim(per) else per
        if not (isinstance(per, (list, tuple, np.ndarray)) and len(per) == 3):
            raise TypeError()

        # Record 1
        values = [
            k,
            nad if nad else "",
            data["density"],
            data["porosity"],
            per[0],
            per[1],
            per[2],
            data["conductivity"],
            data["specific_heat"],
        ]
        out += write_record(values, fmt1)

        # Record 2
        if cond:
            values = [
                data["compressibility"],
                data["expansivity"],
                data["conductivity_dry"],
                data["tortuosity"],
                data["klinkenberg_parameter"],
                data["distribution_coefficient_3"],
                data["distribution_coefficient_4"],
            ]
            out += write_record(values, fmt2)
        else:
            out += write_record([], []) if nad == 2 else []

        # Relative permeability / Capillary pressure
        if nad == 2:
            for key in ["relative_permeability", "capillarity"]:
                if key in data.keys():
                    values = [data[key]["id"], None]
                    values += list(data[key]["parameters"])
                    out += write_record(values, fmt3)
                else:
                    out += write_record([], [])

    return out


@check_parameters(dtypes["MODEL"], keys=("default", "relative_permeability"))
@check_parameters(dtypes["MODEL"], keys=("default", "capillarity"))
@block("RPCAP")
def _write_rpcap(parameters):
    """Write RPCAP block data."""
    data = deepcopy(default)
    data.update(parameters["default"])

    # Formats
    fmt = block_to_format["RPCAP"]
    fmt = str2format(fmt)

    out = []
    for key in ["relative_permeability", "capillarity"]:
        if key in data.keys():
            values = [data[key]["id"], None]
            values += list(data[key]["parameters"])
            out += write_record(values, fmt)
        else:
            out += write_record([], [])

    return out


@check_parameters(dtypes["FLAC"], keys="flac")
@check_parameters(dtypes["MODEL"], keys=("default", "permeability_model"))
@check_parameters(dtypes["MODEL"], keys=("default", "equivalent_pore_pressure"))
@check_parameters(dtypes["MODEL"], keys=("rocks", "permeability_model"), is_list=True)
@check_parameters(
    dtypes["MODEL"], keys=("rocks", "equivalent_pore_pressure"), is_list=True
)
@block("FLAC", multi=True)
def _write_flac(parameters):
    """Write FLAC block data."""
    # Load data
    from ._common import flac

    data = deepcopy(flac)
    data.update(parameters["flac"])

    # Reorder rocks
    if parameters["rocks_order"]:
        order = parameters["rocks_order"]
        for rock in parameters["rocks"].keys():
            if rock not in order:
                order.append(rock)
    else:
        order = parameters["rocks"].keys()

    # Formats
    fmt = block_to_format["FLAC"]
    fmt1 = str2format(fmt[1])
    fmt2 = str2format(fmt[2])
    fmt3 = str2format(fmt[3])

    # Record 1
    values = [
        bool(data["creep"]),
        data["porosity_model"],
        data["version"],
    ]
    out = write_record(values, fmt1)

    # Additional records
    for k in order:
        # Load data
        data = deepcopy(default)
        data.update(parameters["default"])
        data.update(parameters["rocks"][k])

        # Permeability model
        values = [data["permeability_model"]["id"]]
        values += list(data["permeability_model"]["parameters"])
        out += write_record(values, fmt2)

        # Equivalent pore pressure
        values = [data["equivalent_pore_pressure"]["id"], None]
        values += list(data["equivalent_pore_pressure"]["parameters"])
        out += write_record(values, fmt3)

    return out


@check_parameters(dtypes["CHEMP"], keys="chemical_properties", is_list=True)
@block("CHEMP")
def _write_chemp(parameters):
    """Write CHEMP block data."""
    # Load data
    from ._common import chemical_properties

    data = parameters["chemical_properties"]

    # Formats
    fmt = block_to_format["CHEMP"]
    fmt1 = str2format(fmt[1])
    fmt2 = str2format(fmt[2])
    fmt3 = str2format(fmt[3])

    # Record 1
    out = write_record([len(data)], fmt1)

    # Record 2
    for k, v in data.items():
        vv = deepcopy(chemical_properties)
        vv.update(v)

        out += write_record([k], fmt2)

        values = [
            vv["temperature_crit"],
            vv["pressure_crit"],
            vv["compressibility_crit"],
            vv["pitzer_factor"],
            vv["dipole_moment"],
        ]
        out += write_record(values, fmt3)

        values = [
            vv["boiling_point"],
            vv["vapor_pressure_a"],
            vv["vapor_pressure_b"],
            vv["vapor_pressure_c"],
            vv["vapor_pressure_d"],
        ]
        out += write_record(values, fmt3)

        values = [
            vv["molecular_weight"],
            vv["heat_capacity_a"],
            vv["heat_capacity_b"],
            vv["heat_capacity_c"],
            vv["heat_capacity_d"],
        ]
        out += write_record(values, fmt3)

        values = [
            vv["napl_density_ref"],
            vv["napl_temperature_ref"],
            vv["gas_diffusivity_ref"],
            vv["gas_temperature_ref"],
            vv["exponent"],
        ]
        out += write_record(values, fmt3)

        values = [
            vv["napl_viscosity_a"],
            vv["napl_viscosity_b"],
            vv["napl_viscosity_c"],
            vv["napl_viscosity_d"],
            vv["volume_crit"],
        ]
        out += write_record(values, fmt3)

        values = [
            vv["solubility_a"],
            vv["solubility_b"],
            vv["solubility_c"],
            vv["solubility_d"],
        ]
        out += write_record(values, fmt3)

        values = [
            vv["oc_coeff"],
            vv["oc_fraction"],
            vv["oc_decay"],
        ]
        out += write_record(values, fmt3)

    return out


@block("NCGAS")
def _write_ncgas(parameters):
    """Write NCGAS block data."""
    data = parameters["non_condensible_gas"]

    # Formats
    fmt = block_to_format["NCGAS"]
    fmt1 = str2format(fmt[1])
    fmt2 = str2format(fmt[2])

    # Record 1
    out = write_record([len(data)], fmt1)

    # Record 2
    out += write_record(data, fmt2, multi=True)

    return out


@block("MULTI")
def _write_multi(parameters):
    """Write MULTI block data."""
    from ._common import eos

    if "eos" in parameters and parameters["eos"] == "tmvoc":
        for keyword in {"n_component", "n_phase"}:
            if parameters[keyword] is None:
                raise ValueError(
                    "for 'tmvoc', at least 'n_component' and 'n_phase' must be specified"
                )

    # Formats
    fmt = block_to_format["MULTI"]
    fmt = str2format(fmt)

    values = (
        list(eos[parameters["eos"]])
        if parameters["eos"] and parameters["eos"] != "tmvoc"
        else [0, 0, 0, 6]
    )
    values[0] = parameters["n_component"] if parameters["n_component"] else values[0]
    values[1] = values[0] if parameters["isothermal"] else values[0] + 1
    values[2] = parameters["n_phase"] if parameters["n_phase"] else values[2]

    # Handle diffusion
    if len(parameters["diffusion"]):
        values[3] = 8
        parameters["n_phase"] = values[2]  # Save for later check

    # Number of mass components
    if parameters["n_component_incon"]:
        values.append(parameters["n_component_incon"])

    out = write_record(values, fmt)

    return out


@check_parameters(dtypes["SOLVR"], keys="solver")
@block("SOLVR")
def _write_solvr(parameters):
    """Write SOLVR block data."""
    from ._common import solver

    data = deepcopy(solver)
    data.update(parameters["solver"])

    # Formats
    fmt = block_to_format["SOLVR"]
    fmt = str2format(fmt)

    values = [
        data["method"],
        None,
        data["z_precond"],
        None,
        data["o_precond"],
        data["rel_iter_max"],
        data["eps"],
    ]
    out = write_record(values, fmt)

    return out


@block("START")
def _write_start():
    """Write START block data."""
    from ._common import header

    out = "{:5}{}\n".format("----*", header)
    return [out[:11] + "MOP: 123456789*123456789*1234" + out[40:]]


@check_parameters(dtypes["PARAM"], keys="options")
@check_parameters(dtypes["MOP"], keys="extra_options")
@block("PARAM")
def _write_param(parameters, eos_=None):
    """Write PARAM block data."""
    # Load data
    from ._common import extra_options, options

    data = deepcopy(options)
    data.update(parameters["options"])

    # Table
    if not isinstance(data["t_steps"], (list, tuple, np.ndarray)):
        data["t_steps"] = [data["t_steps"]]

    # Formats
    fmt = block_to_format["PARAM"]
    fmt1 = str2format(fmt[1])
    fmt2 = str2format(fmt[2])
    fmt3 = str2format(fmt[3])
    fmt4 = str2format(fmt[4])
    fmt5 = str2format(fmt[5])

    # Record 1
    _mop = deepcopy(extra_options)
    _mop.update(parameters["extra_options"])
    mop = [" " if _mop[k] is None else str(_mop[k]) for k in sorted(_mop.keys())]

    values = [
        data["n_iteration"],
        data["verbosity"],
        data["n_cycle"],
        data["n_second"],
        data["n_cycle_print"],
        "{}".format("".join(mop)),
        None,
        data["temperature_dependence_gas"],
        data["effective_strength_vapor"],
    ]
    out = write_record(values, fmt1)

    # Record 2
    values = [
        data["t_ini"],
        data["t_max"],
        -((len(data["t_steps"]) - 1) // 8 + 1),
        data["t_step_max"],
        None,
        data["gravity"],
        data["t_reduce_factor"],
        data["mesh_scale_factor"],
    ]
    out += write_record(values, fmt2)

    # Record 2.1
    values = [x for x in data["t_steps"]]
    out += write_record(values, fmt3, multi=True)

    # Record 3
    values = [
        data["eps1"],
        data["eps2"],
        None,
        data["w_upstream"],
        data["w_newton"],
        data["derivative_factor"],
    ]
    out += write_record(values, fmt4)

    # Record 4 (TMVOC)
    if eos_ == "tmvoc":
        out += write_record(
            [parameters["default"]["phase_composition"]], str2format("5d")
        )

    # Record 4
    n = min(4, len(parameters["default"]["initial_condition"]))
    values = parameters["default"]["initial_condition"][:n]
    out += write_record(values, fmt5)

    # Record 5 (EOS7R)
    if len(parameters["default"]["initial_condition"]) > 4:
        values = parameters["default"]["initial_condition"][n:]
        out += write_record(values, fmt5)

    return out


@check_parameters(dtypes["SELEC"], keys="selections")
@block("SELEC")
def _write_selec(parameters):
    """Write SELEC block data."""
    # Load data
    from ._common import selections

    data = deepcopy(selections)
    if parameters["selections"]["integers"]:
        data["integers"].update(parameters["selections"]["integers"])
    if "floats" in parameters["selections"] and len(parameters["selections"]["floats"]):
        data["floats"] = parameters["selections"]["floats"]

    # Check floats and overwrite IE(1)
    if data["floats"] is not None and len(data["floats"]):
        if isinstance(data["floats"][0], (list, tuple, np.ndarray)):
            for x in data["floats"]:
                if len(x) > 8:
                    raise ValueError()

            data["integers"][1] = len(data["floats"])
            ndim = 2

        else:
            if len(data["floats"]) > 8:
                raise ValueError()

            data["integers"][1] = 1
            ndim = 1
    else:
        ndim = None

    # Formats
    fmt = block_to_format["SELEC"]
    fmt1 = str2format(fmt[1])
    fmt2 = str2format(fmt[2])

    # Record 1
    values = [data["integers"][k] for k in sorted(data["integers"].keys())]
    out = write_record(values, fmt1)

    # Record 2
    if ndim == 1:
        out += write_record(data["floats"], fmt2)
    elif ndim == 2:
        for x in data["floats"]:
            out += write_record(x, fmt2)

    return out


@block("INDOM", multi=True)
def _write_indom(parameters, eos_):
    """Write INDOM block data."""
    if parameters["rocks_order"]:
        order = parameters["rocks_order"]
        for rock in parameters["rocks"].keys():
            if rock not in order:
                order.append(rock)
    else:
        order = parameters["rocks"].keys()

    # Formats
    fmt = block_to_format["INDOM"]
    fmt1 = str2format(fmt[5])
    fmt2 = str2format(fmt[0])

    out = []
    for k in order:
        cond1 = "initial_condition" in parameters["rocks"][k]
        cond2 = (
            eos_ == "tmvoc"
            and "phase_composition" in parameters["rocks"][k]
            and parameters["rocks"][k]["phase_composition"] is not None
        )

        if cond1 or cond2:
            # Record 1
            values = [k]

            if eos_ == "tmvoc":
                values.append(parameters["rocks"][k]["phase_composition"])

            else:
                values.append(None)

            out += write_record(values, fmt1)

            if cond1:
                data = parameters["rocks"][k]["initial_condition"]

                # Record 2
                n = min(4, len(data))
                values = list(data[:n])
                out += write_record(values, fmt2)

                # Record 3
                if len(data) > 4:
                    values = list(data[4:])
                    out += write_record(values, fmt2)

            else:
                out += ["\n"]

    return out


@check_parameters(dtypes["MOMOP"], keys="more_options")
@block("MOMOP")
def _write_momop(parameters):
    """Write MOMOP block data."""
    from ._common import more_options

    # Formats
    fmt = block_to_format["MOMOP"]
    fmt = str2format(fmt)

    _momop = deepcopy(more_options)
    _momop.update(parameters["more_options"])

    tmp = [" " if _momop[k] is None else str(_momop[k]) for k in sorted(_momop.keys())]
    out = write_record(["".join(tmp)], fmt)

    return out


@block("TIMES")
def _write_times(parameters):
    """Write TIMES block data."""
    data = parameters["times"]
    data = data if np.ndim(data) else [data]

    # Formats
    fmt = block_to_format["TIMES"]
    fmt1 = str2format(fmt[1])
    fmt2 = str2format(fmt[2])

    # Record 1
    out = write_record([len(data)], fmt1)

    # Record 2
    out += write_record(data, fmt2, multi=True)

    return out


@block("FOFT", multi=True)
def _write_foft(parameters):
    """Write FOFT block data."""
    # Formats
    fmt = block_to_format["FOFT"]
    fmt = str2format(fmt[5])

    values = [x for x in parameters["element_history"]]
    out = write_record(values, fmt, multi=True)

    return out


@block("COFT", multi=True)
def _write_coft(parameters):
    """Write COFT block data."""
    # Format
    fmt = block_to_format["COFT"]
    fmt = str2format(fmt[5])

    values = [x for x in parameters["connection_history"]]
    out = write_record(values, fmt, multi=True)

    return out


@block("GOFT", multi=True)
def _write_goft(parameters):
    """Write GOFT block data."""
    # Format
    fmt = block_to_format["GOFT"]
    fmt = str2format(fmt[5])

    values = [x for x in parameters["generator_history"]]
    out = write_record(values, fmt, multi=True)

    return out


@check_parameters(dtypes["GENER"], keys="generators", is_list=True)
@block("GENER", multi=True)
def _write_gener(parameters):
    """Write GENER block data."""
    from ._common import generators

    # Format
    label_length = max(
        [
            len(generator["label"]) if "label" in generator else 0
            for generator in parameters["generators"]
        ]
    )
    label_length = max(label_length, 5)
    fmt = block_to_format["GENER"]
    fmt1 = str2format(fmt[label_length])
    fmt2 = str2format(fmt[0])

    out = []
    for v in parameters["generators"]:
        # Load data
        data = deepcopy(generators)
        data.update(v)

        # Table
        ltab = 1
        if data["times"] is not None and isinstance(
            data["times"], (list, tuple, np.ndarray)
        ):
            ltab = len(data["times"])

            for key in ["rates", "specific_enthalpy"]:
                if data[key] is not None:
                    if ltab == 1 and np.ndim(data[key]) == 1:
                        if len(data[key]) > 1:
                            raise ValueError()

                        data[key] = data[key][0]

                    else:
                        if np.ndim(data[key]) != 1:
                            raise TypeError()

                        if ltab != len(data[key]):
                            raise ValueError()

        else:
            for key in ["rates", "specific_enthalpy"]:
                if key in data and np.ndim(data[key]) > 0:
                    if len(data[key]) > 1:
                        raise ValueError()

                    data[key] = data[key][0]

        itab = (
            "1"
            if isinstance(data["specific_enthalpy"], (list, tuple, np.ndarray))
            else None
        )

        # Record 1
        values = [
            data["label"] if "label" in data else "",
            data["name"],
            data["nseq"],
            data["nadd"],
            data["nads"],
            ltab if ltab > 1 else None,
            None,
            data["type"],
            itab,
            None if ltab > 1 else data["rates"],
            None if ltab > 1 else data["specific_enthalpy"],
            data["layer_thickness"],
        ]
        out += write_record(values, fmt1)

        # Record 2
        out += write_record(data["times"], fmt2, multi=True) if ltab > 1 else []

        # Record 3
        out += write_record(data["rates"], fmt2, multi=True) if ltab > 1 else []

        # Record 4
        if ltab > 1 and data["specific_enthalpy"] is not None:
            if isinstance(data["specific_enthalpy"], (list, tuple, np.ndarray)):
                specific_enthalpy = data["specific_enthalpy"]
            else:
                specific_enthalpy = np.full(ltab, data["specific_enthalpy"])

            out += write_record(specific_enthalpy, fmt2, multi=True)

    return out


@block("DIFFU")
def _write_diffu(parameters):
    """Write DIFFU block data."""
    # Format
    fmt = block_to_format["DIFFU"]
    fmt = str2format(fmt)

    out = []
    for mass in parameters["diffusion"]:
        out += write_record(mass, fmt, multi=True)

    return out


@check_parameters(dtypes["OUTPU"], keys="output")
@block("OUTPU")
def _write_outpu(parameters):
    """Write OUTPU block data."""
    # Load data
    from ._common import output

    data = deepcopy(output)
    data.update(parameters["output"])

    # Format
    fmt = block_to_format["OUTPU"]
    fmt1 = str2format(fmt[1])
    fmt2 = str2format(fmt[2])
    fmt3 = str2format(fmt[3])

    out = []

    # Output format
    out += write_record([data["format"].upper()], fmt1) if data["format"] else []

    # Variables
    if data["variables"]:
        buffer = []
        num_vars = len(data["variables"])

        for variable in data["variables"]:
            values = [variable["name"].upper() if "name" in variable else None]

            if "options" in variable:
                v = variable["options"]

                if np.ndim(v) == 0:
                    values += [v]
                    buffer += write_record(values, fmt3)

                else:
                    if np.ndim(v[0]) == 0:
                        values += list(v)
                        buffer += write_record(values, fmt3)

                    else:
                        for vv in v:
                            values_in = values + list(vv)
                            buffer += write_record(values_in, fmt3)

            else:
                buffer += write_record(values, fmt3)

        out += write_record([str(num_vars)], fmt2)
        out += buffer

    return out


@check_parameters(dtypes["ELEME"], keys="elements", is_list=True)
@block("ELEME", multi=True)
def _write_eleme(parameters):
    """Write ELEME block data."""
    from ._common import elements

    # Reorder elements
    if parameters["elements_order"] is not None:
        order = parameters["elements_order"]
    else:
        order = parameters["elements"].keys()

    # Format
    label_length = len(max(parameters["elements"], key=len))
    fmt = block_to_format["ELEME"]
    fmt = str2format(fmt[label_length])

    out = []
    for k in order:
        data = deepcopy(elements)
        data.update(parameters["elements"][k])

        material = (
            "{:>5}".format(data["material"])
            if isinstance(data["material"], int)
            else data["material"]
        )
        values = [
            k,
            data["nseq"],
            data["nadd"],
            material,
            data["volume"],
            data["heat_exchange_area"],
            data["permeability_modifier"],
            data["center"][0],
            data["center"][1],
            data["center"][2],
        ]
        out += write_record(values, fmt)

    return out


@block("COORD", multi=True)
def _write_coord(parameters):
    """Write COORD block data."""
    # Reorder elements
    if parameters["elements_order"] is not None:
        order = parameters["elements_order"]
    else:
        order = parameters["elements"].keys()

    # Format
    fmt = block_to_format["COORD"]
    fmt = str2format(fmt)

    out = []
    for k in order:
        values = parameters["elements"][k]["center"]
        out += write_record(values, fmt)

    return out


@check_parameters(dtypes["CONNE"], keys="connections", is_list=True)
@block("CONNE", multi=True)
def _write_conne(parameters):
    """Write CONNE block data."""
    from ._common import connections

    # Reorder connections
    if parameters["connections_order"] is not None:
        order = parameters["connections_order"]
    else:
        order = parameters["connections"].keys()

    # Format
    label_length = len(max(parameters["connections"], key=len)) // 2
    fmt = block_to_format["CONNE"]
    fmt = str2format(fmt[label_length])

    out = []
    for k in order:
        data = deepcopy(connections)
        data.update(parameters["connections"][k])

        values = [
            k,
            data["nseq"],
            data["nadd"][0] if data["nadd"] is not None else None,
            data["nadd"][1] if data["nadd"] is not None else None,
            data["permeability_direction"],
            data["nodal_distances"][0],
            data["nodal_distances"][1],
            data["interface_area"],
            data["gravity_cosine_angle"],
            data["radiant_emittance_factor"],
        ]
        out += write_record(values, fmt)

    return out


@check_parameters(dtypes["INCON"], keys="initial_conditions", is_list=True)
@block("INCON", multi=True)
def _write_incon(parameters, eos_=None):
    """Write INCON block data."""
    from ._common import initial_conditions

    # Reorder connections
    if parameters["initial_conditions_order"] is not None:
        order = parameters["initial_conditions_order"]
    else:
        order = parameters["initial_conditions"].keys()

    # Format
    label_length = len(max(parameters["initial_conditions"], key=len))
    fmt = block_to_format["INCON"]
    fmt1 = str2format(
        fmt[eos_][label_length] if eos_ in fmt else fmt["default"][label_length]
    )
    fmt2 = str2format(fmt[0])

    out = []
    for k in order:
        data = deepcopy(initial_conditions)
        data.update(parameters["initial_conditions"][k])

        # Record 1
        values = [
            k,
            None,
            None,
            data["porosity"],
        ]

        if eos_ == "tmvoc":
            values += [data["phase_composition"]]

        else:
            values += list(data["userx"])

        out += write_record(values, fmt1)

        # Record 2
        n = min(4, len(data["values"]))
        out += write_record(data["values"][:n], fmt2)

        # Record 3 (EOS7R)
        if len(data["values"]) > 4:
            out += write_record(data["values"][4:], fmt2)

    return out


@check_parameters(dtypes["MESHM"], keys="meshmaker")
@block("MESHM", multi=True)
def _write_meshm(parameters):
    """Write MESHM block data."""
    from ._common import meshmaker, rz2d, xyz

    data = deepcopy(meshmaker)
    data.update(parameters["meshmaker"])

    # Format
    fmt = block_to_format["MESHM"]
    fmt1 = str2format(fmt[1])

    # Mesh type
    mesh_type = data["type"].upper() if data["type"] else data["type"]
    out = write_record([mesh_type], fmt1)

    # XYZ
    if mesh_type == "XYZ":
        fmt = fmt["XYZ"]
        fmt1 = str2format(fmt[1])
        fmt2 = str2format(fmt[2])
        fmt3 = str2format(fmt[3])

        # Record 1
        out += write_record([data["angle"]], fmt1)

        # Record 2
        for parameter_ in data["parameters"]:
            parameter = deepcopy(xyz)
            parameter.update(parameter_)

            values = [parameter["type"].upper()]
            ndim = np.ndim(parameter["sizes"])

            if ndim == 0:
                values += [
                    parameter["n_increment"],
                    parameter["sizes"],
                ]
                out += write_record(values, fmt2)

            elif ndim == 1:
                values += [
                    parameter["n_increment"]
                    if parameter["n_increment"]
                    else len(parameter["sizes"])
                ]
                out += write_record(values, fmt2)
                out += write_record(parameter["sizes"], fmt3, multi=True)

            else:
                raise ValueError()

    # RZ2D
    elif mesh_type in {"RZ2D", "RZ2DL"}:
        fmt = fmt["RZ2D"]
        fmt0 = str2format(fmt[1])

        for parameter_ in data["parameters"]:
            parameter = deepcopy(rz2d)
            parameter.update(parameter_)

            parameter_type = parameter["type"].upper()
            out += write_record([parameter_type], fmt0)

            if parameter_type == "RADII":
                fmt1 = str2format(fmt["RADII"][1])
                fmt2 = str2format(fmt["RADII"][2])

                out += write_record([len(parameter["radii"])], fmt1)
                out += write_record(parameter["radii"], fmt2, multi=True)

            elif parameter_type == "EQUID":
                fmt1 = str2format(fmt["EQUID"])

                values = [
                    parameter["n_increment"],
                    None,
                    parameter["size"],
                ]
                out += write_record(values, fmt1)

            elif parameter_type == "LOGAR":
                fmt1 = str2format(fmt["LOGAR"])

                values = [
                    parameter["n_increment"],
                    None,
                    parameter["radius"],
                    parameter["radius_ref"],
                ]
                out += write_record(values, fmt1)

            elif parameter_type == "LAYER":
                fmt1 = str2format(fmt["LAYER"][1])
                fmt2 = str2format(fmt["LAYER"][2])

                out += write_record([len(parameter["thicknesses"])], fmt1)
                out += write_record(parameter["thicknesses"], fmt2, multi=True)

    return out


@block("NOVER")
def _write_nover():
    """Write NOVER block data."""
    return []


@block("ENDCY", noend=True)
def _write_endcy():
    """Write ENDCY block data."""
    return []



from __future__ import division

import logging
from copy import deepcopy
from functools import wraps

import numpy as np

dtypes = {
    "PARAMETERS": {
        "title": "str_array_like",
        "eos": "str",
        "n_component": "int",
        "n_phase": "int",
        "n_component_incon": "int",
        "flac": "dict",
        "chemical_properties": "dict",
        "non_condensible_gas": "str_array_like",
        "isothermal": "bool",
        "start": "bool",
        "nover": "bool",
        "rocks": "dict",
        "rocks_order": "array_like",
        "options": "dict",
        "extra_options": "dict",
        "more_options": "dict",
        "selections": "dict",
        "solver": "dict",
        "generators": "array_like",
        "times": "array_like",
        "element_history": "array_like",
        "connection_history": "array_like",
        "generator_history": "array_like",
        "diffusion": "array_like",
        "output": "dict",
        "elements": "dict",
        "elements_order": "array_like",
        "coordinates": "bool",
        "connections": "dict",
        "connections_order": "array_like",
        "initial_conditions": "dict",
        "initial_conditions_order": "array_like",
        "meshmaker": "dict",
        "default": "dict",
    },
    "ROCKS": {
        "density": "scalar",
        "porosity": "scalar",
        "permeability": "scalar_array_like",
        "conductivity": "scalar",
        "specific_heat": "scalar",
        "compressibility": "scalar",
        "expansivity": "scalar",
        "conductivity_dry": "scalar",
        "tortuosity": "scalar",
        "klinkenberg_parameter": "scalar",
        "distribution_coefficient_3": "scalar",
        "distribution_coefficient_4": "scalar",
        "initial_condition": "array_like",
        "relative_permeability": "dict",
        "capillarity": "dict",
        "permeability_model": "dict",
        "equivalent_pore_pressure": "dict",
        "phase_composition": "int",
    },
    "FLAC": {"creep": "bool", "porosity_model": "int", "version": "int"},
    "CHEMP": {
        "temperature_crit": "scalar",
        "pressure_crit": "scalar",
        "compressibility_crit": "scalar",
        "pitzer_factor": "scalar",
        "dipole_moment": "scalar",
        "boiling_point": "scalar",
        "vapor_pressure_a": "scalar",
        "vapor_pressure_b": "scalar",
        "vapor_pressure_c": "scalar",
        "vapor_pressure_d": "scalar",
        "molecular_weight": "scalar",
        "heat_capacity_a": "scalar",
        "heat_capacity_b": "scalar",
        "heat_capacity_c": "scalar",
        "heat_capacity_d": "scalar",
        "napl_density_ref": "scalar",
        "napl_temperature_ref": "scalar",
        "gas_diffusivity_ref": "scalar",
        "gas_temperature_ref": "scalar",
        "exponent": "scalar",
        "napl_viscosity_a": "scalar",
        "napl_viscosity_b": "scalar",
        "napl_viscosity_c": "scalar",
        "napl_viscosity_d": "scalar",
        "volume_crit": "scalar",
        "solubility_a": "scalar",
        "solubility_b": "scalar",
        "solubility_c": "scalar",
        "solubility_d": "scalar",
        "oc_coeff": "scalar",
        "oc_fraction": "scalar",
        "oc_decay": "scalar",
    },
    "MODEL": {"id": "int", "parameters": "array_like"},
    "PARAM": {
        "n_iteration": "int",
        "n_cycle": "int",
        "n_second": "int",
        "n_cycle_print": "int",
        "verbosity": "int",
        "temperature_dependence_gas": "scalar",
        "effective_strength_vapor": "scalar",
        "t_ini": "scalar",
        "t_max": "scalar",
        "t_steps": "scalar_array_like",
        "t_step_max": "scalar",
        "t_reduce_factor": "scalar",
        "gravity": "scalar",
        "mesh_scale_factor": "scalar",
        "eps1": "scalar",
        "eps2": "scalar",
        "w_upstream": "scalar",
        "w_newton": "scalar",
        "derivative_factor": "scalar",
    },
    "MOP": {i + 1: "int" for i in range(24)},
    "MOMOP": {i + 1: "int" for i in range(40)},
    "SELEC": {"integers": "dict", "floats": "array_like"},
    "SOLVR": {
        "method": "int",
        "z_precond": "str",
        "o_precond": "str",
        "rel_iter_max": "scalar",
        "eps": "scalar",
    },
    "GENER": {
        "label": "str",
        "name": "str",
        "nseq": "scalar",
        "nadd": "scalar",
        "nads": "scalar",
        "type": "str",
        "times": "scalar_array_like",
        "rates": "scalar_array_like",
        "specific_enthalpy": "scalar_array_like",
        "layer_thickness": "scalar",
    },
    "OUTPU": {"format": "str", "variables": "array_like"},
    "ELEME": {
        "nseq": "int",
        "nadd": "int",
        "material": "str_int",
        "volume": "scalar",
        "heat_exchange_area": "scalar",
        "permeability_modifier": "scalar",
        "center": "array_like",
    },
    "CONNE": {
        "nseq": "int",
        "nadd": "array_like",
        "permeability_direction": "int",
        "nodal_distances": "array_like",
        "interface_area": "scalar",
        "gravity_cosine_angle": "scalar",
        "radiant_emittance_factor": "scalar",
    },
    "INCON": {
        "porosity": "scalar",
        "userx": "array_like",
        "values": "array_like",
        "phase_composition": "int",
    },
    "MESHM": {"type": "str", "parameters": "array_like", "angle": "scalar"},
}


str_to_dtype = {
    "int": (int, np.int32, np.int64),
    "float": (float, np.float32, np.float64),
    "str": (str,),
    "bool": (bool,),
    "str_int": (str, int, np.int32, np.int64),
    "array_like": (list, tuple, np.ndarray),
    "dict": (dict,),
    "scalar": (int, float, np.int32, np.int64, np.float32, np.float64),
    "scalar_array_like": (
        int,
        float,
        list,
        tuple,
        np.int32,
        np.int64,
        np.float32,
        np.float64,
        np.ndarray,
    ),
    "str_array_like": (str, list, tuple, np.ndarray),
}


def block(keyword, multi=False, noend=False):
    """Decorate block writing functions."""

    def decorator(func):
        from ._common import header

        @wraps(func)
        def wrapper(*args, **kwargs):
            head_fmt = "{:5}{}" if noend else "{:5}{}\n"
            out = [head_fmt.format(keyword, header)]
            out += func(*args, **kwargs)
            out += ["\n"] if multi else []

            return out

        return wrapper

    return decorator


def check_parameters(input_types, keys=None, is_list=False):
    """Decorate function to check input parameters."""

    def _check_parameters(params, keys=None):
        for k, v in params.items():
            # Check whether parameters contain unknown keys
            # Log error if it does and skip
            if k not in input_types.keys():
                logging.warning(
                    "Unknown key '{}'{}. Skipping.".format(
                        k, " in {}".format(keys) if keys else ""
                    )
                )
                continue

            # Check input types
            input_type = str_to_dtype[input_types[k]]
            if not (v is None or isinstance(v, input_type)):
                raise TypeError(
                    "Invalid type for parameter '{}' {}(expected {}).".format(
                        k, "in {} ".format(keys) if keys else "", input_types[k],
                    )
                )

    keys = [keys] if isinstance(keys, str) else keys

    def decorator(func, *args, **kwargs):
        @wraps(func)
        def wrapper(parameters, *args, **kwargs):
            if not keys:
                _check_parameters(parameters)

            else:
                params = deepcopy(parameters[keys[0]])
                keys_str = "['{}']".format(keys[0])
                if is_list:
                    if isinstance(params, dict):
                        for k, v in params.items():
                            tmp = keys_str
                            tmp += "['{}']".format(k)

                            try:
                                for key in keys[1:]:
                                    v = v[key]
                                    tmp += "['{}']".format(key)
                                _check_parameters(v, tmp)

                            except KeyError:
                                continue
                    else:
                        for i, param in enumerate(params):
                            tmp = "{}[{}]".format(keys_str, i)
                            _check_parameters(param, tmp)

                else:
                    for key in keys[1:]:
                        params = params[key]
                        keys_str += "['{}']".format(key)

                    _check_parameters(params, keys_str)

            out = func(parameters, *args, **kwargs)

            return out

        return wrapper

    return decorator


def read_record(data, fmt):
    """Parse string to data given format."""
    token_to_type = {
        "s": str,
        "S": str,
        "d": int,
        "f": str2float,
        "e": str2float,
    }

    i = 0
    out = []
    for token in fmt.split(","):
        n = int(token[:-1].split(".")[0])
        tmp = data[i : i + n]
        tmp = tmp if token[-1] == "S" else tmp.strip()
        out.append(token_to_type[token[-1]](tmp) if tmp else None)
        i += n

    return out


def write_record(data, fmt, multi=False):
    """Return a list of record strings given format."""

    def to_str(x, fmt):
        x = "" if x is None else x

        if not isinstance(x, str):
            # Special handling for floating point numbers
            if "f" in fmt:
                # Number of decimals is specified
                if "." in fmt:
                    n = int(fmt[3:].split(".")[0])
                    tmp = fmt.format(x)

                    if len(tmp) > n:
                        return fmt.replace("f", "e").format(x)

                    else:
                        return tmp

                # Let Python decides the format
                else:
                    n = int(fmt[3:].split("f")[0])
                    tmp = str(float(x))

                    if len(tmp) > n:
                        fmt = "{{:>{}.{}e}}".format(n, n - 7)

                        return fmt.format(x)

                    else:
                        fmt = "{{:>{}}}".format(n)

                        return fmt.format(tmp)

            else:
                return fmt.format(x)

        else:
            return fmt.replace("g", "").replace("e", "").replace("f", "").format(x)

    if not multi:
        data = [to_str(d, f) for d, f in zip(data, fmt)]
        out = ["{:80}\n".format("".join(data))]

    else:
        n = len(data)
        ncol = len(fmt)
        data = [
            data[ncol * i : min(ncol * i + ncol, n)]
            for i in range(int(np.ceil(n / ncol)))
        ]

        out = []
        for d in data:
            d = [to_str(dd, f) for dd, f in zip(d, fmt)]
            out += ["{:80}\n".format("".join(d))]

    return out


def str2float(s):
    """Convert variable string to float."""
    try:
        return float(s)
    except ValueError:
        # It's probably something like "0.0001-001"
        significand, exponent = s[:-4], s[-4:]
        return float("{}e{}".format(significand, exponent))


def prune_nones_dict(data):
    """Remove None key/value pairs from dict."""
    return {k: v for k, v in data.items() if v is not None}


def prune_nones_list(data):
    """Remove trailing None values from list."""
    return [x for i, x in enumerate(data) if any(xx is not None for xx in data[i:])]



from ._corey import Corey
from ._fatt_klikoff import FattKlikoff
from ._grant import Grant
from ._linear import Linear
from ._pickens import Pickens
from ._van_genuchten_mualem import vanGenuchtenMualem
from ._verma import Verma

__all__ = [
    "Linear",
    "Pickens",
    "Corey",
    "Grant",
    "FattKlikoff",
    "vanGenuchtenMualem",
    "Verma",
]



from ._base import BaseRelativePermeability

__all__ = [
    "vanGenuchtenMualem",
]


class vanGenuchtenMualem(BaseRelativePermeability):

    _id = 7
    _name = "van Genuchten-Mualem"

    def __init__(self, m, slr, sls, sgr):
        """
        Van Genuchten-Mualem function.

        After Mualem (1976) and van Genuchten (1980).

        Parameters
        ----------
        m : scalar
            Related to pore size distribution (RP(1)).
        slr : scalar
            Irreducible liquid saturation (RP(2)).
        sls : scalar
            Maximum liquid saturation (RP(3)).
        sgr : scalar
            Irreducible gas saturation (RP(4)).

        """
        self.parameters = [m, slr, sls, sgr]

    def _eval(self, sl, *args):
        """Van Genuchten-Mualem function."""
        m, slr, sls, sgr = args
        if sl >= sls:
            kl = 1.0
            kg = 0.0

        else:
            Seff = (sl - slr) / (sls - slr)
            kl = (
                Seff ** 0.5 * (1.0 - (1.0 - Seff ** (1.0 / m)) ** m) ** 2
                if Seff > 0.0
                else 0.0
            )

            Shat = (sl - slr) / (1.0 - slr - sgr)
            Shat = max(Shat, 0.0)
            Shat = min(Shat, 1.0)
            kg = 1.0 - kl if sgr <= 0.0 else (1.0 - Shat ** 2) * (1.0 - Shat) ** 2

        return kl, kg

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._m, self._slr, self._sls, self._sgr]

    @parameters.setter
    def parameters(self, value):
        if len(value) != 4:
            raise ValueError()
        self._m = value[0]
        self._slr = value[1]
        self._sls = value[2]
        self._sgr = value[3]



import glob
import os

import helpers
import numpy as np
import pytest

import toughio


@pytest.mark.parametrize("dirname", [helpers.tempdir(), "abc"])
def test_co2tab(dirname):
    argv = [dirname]

    if os.path.isdir(dirname):
        toughio._cli.co2tab(argv)
        assert os.path.isfile(os.path.join(dirname, "CO2TAB"))
    else:
        with pytest.raises(ValueError):
            toughio._cli.co2tab(argv)


@pytest.mark.parametrize(
    "filename, mesh, voxelize, ext",
    [
        ("OUTPUT_ELEME.csv", True, False, "vtu"),
        ("OUTPUT_ELEME.csv", False, False, "vtu"),
        ("OUTPUT_ELEME.tec", False, False, "vtu"),
        ("OUTPUT_ELEME.csv", True, False, "xdmf"),
        ("OUTPUT_ELEME.csv", False, False, "xdmf"),
        ("OUTPUT_ELEME.tec", False, False, "xdmf"),
        ("OUTPUT_ELEME.csv", True, True, "vtu"),
        ("OUTPUT_ELEME.csv", False, True, "vtu"),
        ("OUTPUT_ELEME.tec", False, True, "vtu"),
        ("OUTPUT_ELEME.csv", True, True, "xdmf"),
        ("OUTPUT_ELEME.csv", False, True, "xdmf"),
        ("OUTPUT_ELEME.tec", False, True, "xdmf"),
    ],
)
def test_export(filename, mesh, voxelize, ext):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(this_dir, "support_files", "outputs", filename)

    outputs = toughio.read_output(filename)

    output_filename = "{}.{}".format(helpers.tempdir(helpers.random_string(10)), ext)
    argv = [
        filename,
        "-o",
        output_filename,
        "-f",
        ext,
    ]

    if voxelize:
        argv += [
            "-v",
            "--origin",
            "0.0",
            "0.0",
            "-50.5",
            "--layer",  # Test files' mesh is generated by layers
        ]
    elif mesh:
        argv += [
            "-m",
            os.path.join(this_dir, "support_files", "outputs", "mesh.pickle"),
        ]

    if ext != "xdmf":
        t = np.random.randint(len(outputs))
        argv += ["-t", str(t)]
        output = outputs[t]

    toughio._cli.export(argv)

    if ext != "xdmf":
        mesh_in = toughio.read_mesh(output_filename)

        centers_ref = np.column_stack([output.data[dim] for dim in ["X", "Y", "Z"]])
        assert (
            np.allclose(centers_ref, mesh_in.centers)
            if mesh or voxelize
            else np.allclose(centers_ref[:, :2], mesh_in.points[:, :2])
        )

        for k, v in output.data.items():
            if k not in {"X", "Y", "Z"}:
                assert (
                    np.allclose(v, mesh_in.cell_data[k])
                    if mesh or voxelize
                    else np.allclose(v, mesh_in.point_data[k])
                )
    else:
        mesh_in = toughio.read_time_series(output_filename)
        points, cells, point_data, cell_data, time_steps = mesh_in

        centers_ref = np.column_stack(
            [outputs[-1].data[dim] for dim in ["X", "Y", "Z"]]
        )
        assert (
            np.allclose(
                centers_ref,
                np.concatenate([points[c.data].mean(axis=1) for c in cells]),
            )
            if mesh or voxelize
            else np.allclose(centers_ref[:, :2], points[:, :2])
        )

        assert len(point_data) == len(outputs)
        assert len(cell_data) == len(outputs)

        assert (
            all(pdata == {} for pdata in point_data)
            if mesh or voxelize
            else all(cdata == {} for cdata in cell_data)
        )

        for t, pdata in enumerate(point_data):
            for k, v in pdata.items():
                assert np.allclose(v, outputs[t].data[k])

        for t, cdata in enumerate(cell_data):
            for k, v in cdata.items():
                assert np.allclose(v, outputs[t].data[k])

        time_steps_ref = [output.time for output in outputs]
        assert np.allclose(time_steps, time_steps_ref)


@pytest.mark.parametrize(
    "file_format, split, connection",
    [
        ("csv", True, False),
        ("csv", True, True),
        ("csv", False, False),
        ("csv", False, True),
        ("tecplot", True, False),
        ("tecplot", False, False),
    ],
)
def test_extract(file_format, split, connection):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(this_dir, "support_files", "outputs", "OUTPUT.out")
    mesh_file = os.path.join(this_dir, "support_files", "outputs", "MESH.out")

    base_filename = "OUTPUT_ELEME" if not connection else "OUTPUT_CONNE"

    tempdir = helpers.tempdir()
    output_filename = os.path.join(tempdir, "{}.csv".format(base_filename))

    argv = [
        filename,
        mesh_file,
        "-o",
        output_filename,
        "-f",
        file_format,
    ]
    argv += ["--split"] if split else []
    argv += ["--connection"] if connection else []
    toughio._cli.extract(argv)

    filename_ref = os.path.join(
        this_dir, "support_files", "outputs", "{}.csv".format(base_filename)
    )
    outputs_ref = toughio.read_output(filename_ref)

    if not split:
        outputs = toughio.read_output(output_filename, connection=connection)

        for output_ref, output in zip(outputs_ref, outputs):
            assert output_ref.time == output.time
            for k, v in output_ref.data.items():
                assert np.allclose(v.mean(), output.data[k].mean(), atol=1.0e-2)
    else:
        filenames = glob.glob(os.path.join(tempdir, "{}_*.csv".format(base_filename)))
        for i, output_filename in enumerate(sorted(filenames)):
            outputs = toughio.read_output(output_filename)

            assert len(outputs) == 1

            output = outputs[0]
            output_ref = outputs_ref[i]

            assert output_ref.time == output.time
            for k, v in output_ref.data.items():
                assert np.allclose(v.mean(), output.data[k].mean(), atol=1.0e-2)


@pytest.mark.parametrize("incon", [True, False])
def test_merge(incon):
    tempdir = helpers.tempdir()
    filename = os.path.join(tempdir, "INFILE")
    mesh_file = os.path.join(tempdir, "MESH")
    output_filename = os.path.join(tempdir, "OUTFILE")

    with open(filename, "w") as f:
        n_lines_rocks, n_lines_param = np.random.randint(20, size=2) + 1

        f.write("ROCKS\n")
        for _ in range(n_lines_rocks):
            f.write("{}\n".format(helpers.random_string(80)))
        f.write("\n")

        f.write("PARAM\n")
        for _ in range(n_lines_param):
            f.write("{}\n".format(helpers.random_string(80)))

        f.write("ENDCY\n")

    with open(mesh_file, "w") as f:
        n_lines_eleme, n_lines_conne = np.random.randint(20, size=2) + 1

        f.write("ELEME\n")
        for _ in range(n_lines_eleme):
            f.write("{}\n".format(helpers.random_string(80)))
        f.write("\n")

        f.write("CONNE\n")
        for _ in range(n_lines_conne):
            f.write("{}\n".format(helpers.random_string(80)))
        f.write("\n")

    if incon:
        incon_file = os.path.join(tempdir, "INCON")

        with open(incon_file, "w") as f:
            n_lines_incon = np.random.randint(20) + 1

            f.write("INCON\n")
            for _ in range(n_lines_incon):
                f.write("{}\n".format(helpers.random_string(80)))
            f.write("\n")

    argv = [filename, output_filename]
    toughio._cli.merge(argv)

    with open(output_filename, "r") as f:
        n_lines = 0
        keywords = []
        for line in f:
            n_lines += 1
            if line[:5] in {"ROCKS", "PARAM", "ENDCY", "ELEME", "CONNE", "INCON"}:
                keywords.append(line[:5])

    n_lines_ref = n_lines_rocks + n_lines_param + n_lines_eleme + n_lines_conne
    n_lines_ref += n_lines_incon + 10 if incon else 8
    assert n_lines == n_lines_ref

    assert (
        keywords == ["ROCKS", "PARAM", "ELEME", "CONNE", "ENDCY"]
        if not incon
        else keywords == ["ROCKS", "PARAM", "ELEME", "CONNE", "INCON", "ENDCY"]
    )


@pytest.mark.parametrize("reset", [True, False])
def test_save2incon(reset):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(this_dir, "support_files", "outputs", "SAVE.out")
    save = toughio.read_output(filename)

    output_filename = helpers.tempdir(helpers.random_string(10))
    argv = [
        filename,
        output_filename,
    ]

    if reset:
        argv += ["-r"]

    toughio._cli.save2incon(argv)

    incon = toughio.read_output(output_filename)

    assert save.labels.tolist() == incon.labels.tolist()
    helpers.allclose_dict(save.data, incon.data)



import os

import matplotlib.pyplot as plt
import numpy as np
import pytest

import toughio

if not os.environ.get("DISPLAY", ""):
    plt.switch_backend("Agg")


@pytest.mark.parametrize(
    "model, parameters",
    [
        (toughio.relative_permeability.Linear, [0.2, 0.3, 0.8, 0.9]),
        (toughio.relative_permeability.Pickens, [2.0]),
        (toughio.relative_permeability.Corey, [0.3, 0.05]),
        (toughio.relative_permeability.Grant, [0.3, 0.05]),
        (toughio.relative_permeability.FattKlikoff, [0.3]),
        (toughio.relative_permeability.vanGenuchtenMualem, [0.457, 0.15, 1.0, 0.1]),
        (toughio.relative_permeability.Verma, []),
    ],
)
def test_relative_permeability(model, parameters, monkeypatch):
    import json

    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(
        this_dir, "support_files", "relative_permeability_references.json"
    )
    with open(filename, "r") as f:
        relperm_ref = json.load(f)

    sl = np.linspace(0.0, 1.0, 201)

    perm = model(*parameters)
    relperm = np.transpose(perm(sl))

    assert np.allclose(relperm, relperm_ref[perm.name])

    monkeypatch.setattr(plt, "show", lambda: None)
    perm.plot()

    print(perm)



import numpy as np

from .._mesh import Mesh
from ._structured_grid import structured_grid

__all__ = [
    "cylindric_grid",
]


class CylindricMesh(Mesh):
    def __init__(self, dr, dz, layer, *args, **kwargs):
        """
        Cylindric mesh.

        This class is only intended to be used as output of :func:`cylindric_grid`.

        Note
        ----
        This class inherits from :class:`toughio.Mesh` but overwrites how face areas
        and volumes are calculated.

        """
        super(CylindricMesh, self).__init__(*args, **kwargs)
        self._dr = dr
        self._dz = dz[::-1]  # From top to bottom
        self._layer = layer

    def _get_areas_heights(self):
        """Return areas and heights of cells in mesh."""
        nr, nz = len(self._dr), len(self._dz)
        r2 = np.cumsum(self._dr) ** 2
        areas = np.tile(np.concatenate(([r2[0]], r2[1:] - r2[:-1])), nz) * np.pi
        heights = np.tile(self._dz[:, None], nr).ravel()

        return areas, heights

    @property
    def face_areas(self):
        """Areas of faces in mesh."""
        nr, nz = len(self._dr), len(self._dz)
        dr = np.concatenate(([0.0], self._dr))
        perimeters_in = np.tile(np.cumsum(dr[:-1]), nz) * 2.0 * np.pi
        perimeters_out = np.tile(np.cumsum(dr[1:]), nz) * 2.0 * np.pi
        areas, heights = self._get_areas_heights()
        sections = np.tile(self._dr, nz) * heights
        out = np.transpose(
            [
                areas,
                areas,
                sections,
                perimeters_out * heights,
                sections,
                perimeters_in * heights,
            ]
        )

        return (
            out
            if self._layer
            else out.reshape((nz, nr, 6)).reshape((nz * nr, 6), order="F")
        )

    @property
    def volumes(self):
        """Volumes of cell in mesh."""
        nr, nz = len(self._dr), len(self._dz)
        areas, heights = self._get_areas_heights()
        out = areas * heights

        return out if self._layer else out.reshape((nz, nr)).ravel(order="F")


def cylindric_grid(dr, dz, origin_z=None, layer=False, material="dfalt"):
    """
    Generate a cylindric mesh as a radial XZ structured grid.

    Parameters
    ----------
    dr : array_like
        Grid spacing along X axis.
    dz : array_like
        Grid spacing along Z axis.
    origin_z : scalar, optional, default None
        Depth of origin point.
    layer : bool, optional, default False
        If `True`, mesh will be generated by layers.
    material : str, optional, default 'dfalt'
        Default material name.

    Returns
    -------
    toughio.Mesh
        Output cylindric mesh.

    """
    if not isinstance(dr, (list, tuple, np.ndarray)):
        raise TypeError()
    if not isinstance(dz, (list, tuple, np.ndarray)):
        raise TypeError()
    if not (origin_z is None or isinstance(origin_z, (int, float))):
        raise TypeError()
    if not isinstance(material, str):
        raise TypeError()

    dr = np.asarray(dr)
    dz = np.asarray(dz)
    if not (dr >= 0.0).all():
        raise ValueError()
    if not (dz >= 0.0).all():
        raise ValueError()
    origin_z = origin_z if origin_z is not None else -dz.sum()

    mesh = structured_grid(
        dr, [1.0], dz, origin=[0.0, -0.5, origin_z], layer=layer, material=material,
    )

    return CylindricMesh(
        dr,
        dz,
        layer,
        points=mesh.points,
        cells=mesh.cells,
        point_data=mesh.point_data,
        cell_data=mesh.cell_data,
        field_data=mesh.field_data,
    )



from __future__ import with_statement

try:
    import cPickle as pickle
except ImportError:
    import pickle

from ..._common import open_file

__all__ = [
    "read",
    "write",
]


def read(filename):
    """
    Import pickled :class:`toughio.Mesh`.

    Parameters
    ----------
    filename : str
        Input file name.

    Returns
    -------
    toughio.Mesh
        Output mesh.

    """
    with open_file(filename, "rb") as f:
        mesh = pickle.load(f)

    return mesh


def write(filename, mesh, protocol=pickle.HIGHEST_PROTOCOL):
    """
    Pickle :class:`toughio.Mesh`.

    Parameters
    ----------
    filename : str
        Output file name.
    mesh : toughio.Mesh
        Mesh to pickle.
    protocol : integer, optional, default `pickle.HIGHEST_PROTOCOL`
        :mod:`pickle` protocol version.

    """
    with open_file(filename, "wb") as f:
        pickle.dump(mesh, f, protocol=protocol)



from ._base import BaseRelativePermeability

__all__ = [
    "Verma",
]


class Verma(BaseRelativePermeability):

    _id = 8
    _name = "Verma"

    def __init__(self, slr=0.2, sls=0.895, a=1.259, b=-1.7615, c=0.5089):
        """
        Verma's function.

        After Verma et al. (1985).

        Parameters
        ----------
        slr : scalar
            Irreducible liquid saturation (RP(1)).
        sls : scalar
            Maximum liquid saturation (RP(2)).
        a : scalar
            A (RP(3)).
        b : scalar
            B (RP(4)).
        c : scalar
            C (RP(5)).

        """
        self.parameters = [slr, sls, a, b, c]

    def _eval(self, sl, *args):
        """Verma's function."""
        slr, sls, a, b, c = args
        Shat = (sl - slr) / (sls - slr)
        Shat = max(Shat, 0.0)
        Shat = min(Shat, 1.0)
        kl = Shat ** 3
        kg = a + b * Shat + c * Shat ** 2
        kg = max(kg, 0.0)
        kg = min(kg, 1.0)
        return kl, kg

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._slr, self._sls, self._a, self._b, self._c]

    @parameters.setter
    def parameters(self, value):
        if len(value) != 5:
            raise ValueError()
        self._slr = value[0]
        self._sls = value[1]
        self._a = value[2]
        self._b = value[3]
        self._c = value[4]



import helpers
import numpy as np
import pytest

import toughio


def test_cylindric_grid():
    dr = np.array([1.0, 2.0, 3.0, 4.0])
    dz = np.array([1.0, 2.0, 3.0])
    mesh = toughio.meshmaker.cylindric_grid(dr, dz)

    perimeters = 2.0 * np.pi * dr.cumsum()
    base_areas = np.pi * dr.cumsum() ** 2
    surface_areas = perimeters * dz.sum()
    section_areas = dr.sum() * dz.sum()

    volumes = mesh.volumes.sum()
    assert np.allclose(volumes, base_areas[-1] * dz.sum())

    face_areas_top = np.array(mesh.face_areas)[:, 0].sum()
    assert np.allclose(face_areas_top, base_areas[-1] * len(dz))

    face_areas_bottom = np.array(mesh.face_areas)[:, 1].sum()
    assert np.allclose(face_areas_bottom, base_areas[-1] * len(dz))

    face_areas_section_1 = np.array(mesh.face_areas)[:, 2].sum()
    assert np.allclose(face_areas_section_1, section_areas)

    face_areas_outer = np.array(mesh.face_areas)[:, 3].sum()
    assert np.allclose(face_areas_outer, surface_areas.sum())

    face_areas_section_2 = np.array(mesh.face_areas)[:, 4].sum()
    assert np.allclose(face_areas_section_2, section_areas)

    face_areas_inner = np.array(mesh.face_areas)[:, 5].sum()
    assert np.allclose(face_areas_inner, surface_areas[:-1].sum())


@pytest.mark.parametrize("ndim", [2, 3])
def test_structured_grid(ndim):
    dx = np.array([1.0, 2.0, 3.0, 4.0])
    dy = np.array([1.0, 2.0, 3.0])
    dz = np.array([1.0, 2.0]) if ndim == 3 else None

    origin = np.random.rand(ndim)
    mesh = toughio.meshmaker.structured_grid(dx, dy, dz, origin=origin)

    assert np.allclose(origin, mesh.points.min(axis=0)[:ndim])

    if ndim == 3:
        volumes_ref = dx.sum() * dy.sum() * dz.sum()
        assert np.allclose(volumes_ref, mesh.volumes.sum())


@pytest.mark.parametrize("ndim", [2, 3])
def test_triangulate(ndim):
    points = np.random.rand(100, ndim)
    mesh = toughio.meshmaker.triangulate(points)

    assert np.allclose(points, mesh.points)
    assert mesh.cells[0].type == ("triangle" if ndim == 2 else "tetra")


def test_voxelize():
    dx = np.array([1.0, 2.0, 3.0, 4.0])
    dy = np.array([1.0, 2.0, 3.0])
    dz = np.array([1.0, 2.0])
    origin = np.random.rand(3)
    mesh_ref = toughio.meshmaker.structured_grid(dx, dy, dz, origin=origin)

    mesh = toughio.meshmaker.voxelize(mesh_ref.centers, origin=origin)
    helpers.allclose_mesh(mesh_ref, mesh)


def test_layer():
    dr = np.array([1.0, 2.0, 3.0, 4.0])
    dz = np.array([1.0, 2.0, 3.0])

    mesh = toughio.meshmaker.cylindric_grid(dr, dz, layer=False)
    mesh_layer = toughio.meshmaker.cylindric_grid(dr, dz, layer=True)

    point = 3.5, 0.0, 1.5
    i1 = mesh.near(point)
    i2 = mesh_layer.near(point)

    assert np.allclose(mesh.volumes[i1], mesh_layer.volumes[i2])
    assert np.allclose(mesh.face_areas[i1], mesh_layer.face_areas[i2])


def test_from_meshmaker_xyz():
    parameters = {
        "meshmaker": {
            "type": "xyz",
            "parameters": [
                {"type": "nx", "n_increment": 3, "sizes": 14.09},
                {"type": "ny", "n_increment": 10, "sizes": np.arange(10) + 1.0},
                {"type": "nz", "n_increment": 3, "sizes": [14.0, 0.9, 19.91]},
            ],
        }
    }
    mesh = toughio.meshmaker.from_meshmaker(parameters)

    assert mesh.n_cells == 90
    assert np.allclose(mesh.points[:, 0].max(), 3 * 14.09)
    assert np.allclose(mesh.points[:, 1].max(), 55.0)
    assert np.allclose(mesh.points[:, 2].min(), -14.0 - 0.9 - 19.91)


def test_from_meshmaker_rz2d():
    parameters = {
        "meshmaker": {
            "type": "rz2d",
            "parameters": [
                {"type": "radii", "radii": np.arange(10)},
                {"type": "equid", "n_increment": 3, "size": 14.09},
                {
                    "type": "logar",
                    "n_increment": 14,
                    "radius": 64.0,
                    "radius_ref": 0.9,
                },
                {"type": "layer", "thicknesses": [14.0, 0.9, 19.91]},
            ],
        }
    }
    mesh = toughio.meshmaker.from_meshmaker(parameters)

    assert mesh.n_cells == 26 * 3
    assert np.allclose(mesh.points[:, 0].max(), 64.0)
    assert np.allclose(mesh.points[:, 2].min(), -14.0 - 0.9 - 19.91)
    assert np.allclose(mesh._dr.sum(), 64.0)
    assert np.allclose(mesh._dz.sum(), 14 + 0.9 + 19.91)



from __future__ import with_statement

import numpy as np

from ...._common import open_file
from ...._mesh.tecplot._tecplot import _read_variables, _read_zone
from .._common import to_output

__all__ = [
    "read",
    "write",
]


def read(filename, file_type, file_format, labels_order):
    """Read OUTPUT_ELEME.tec."""
    with open_file(filename, "r") as f:
        # Look for header (VARIABLES)
        while True:
            line = f.readline().strip()
            if line.upper().startswith("VARIABLES"):
                break

        # Read header (VARIABLES)
        headers = _read_variables(line)

        # Loop until end of file
        times, labels, variables = [], [], []
        line = f.readline().upper().strip()
        while True:
            # Read zone
            if line.startswith("ZONE"):
                zone = _read_zone(line)
                zone["T"] = float(zone["T"].split()[0]) if "T" in zone.keys() else None
                if "I" not in zone.keys():
                    raise ValueError()

            # Read data
            data = []
            for _ in range(zone["I"]):
                line = f.readline().strip()
                data.append([float(x) for x in line.split()])
            data = np.array(data)

            # Output
            times.append(zone["T"])
            labels.append(None)
            variables.append(data)

            line = f.readline().upper().strip()
            if not line:
                break

    return to_output(
        file_type, file_format, labels_order, headers, times, labels, variables
    )


def write(filename, output):
    """Write OUTPUT_ELEME.tec."""
    out = output[-1]
    headers = []
    if "X" in out.data.keys():
        headers += ["X"]
    else:
        raise ValueError()
    if "Y" in out.data.keys():
        headers += ["Y"]
    else:
        raise ValueError()
    headers += ["Z"] if "Z" in out.data.keys() else []
    headers += [k for k in out.data.keys() if k not in {"X", "Y", "Z"}]

    with open_file(filename, "w") as f:
        # Headers
        record = "".join("{:>18}".format(header) for header in headers)
        f.write("{}{}\n".format("{:>18}".format("VARIABLES       ="), record))

        # Data
        for out in output:
            # Zone
            record = ' ZONE T="{:14.7e} SEC"  I = {:8d}'.format(
                out.time, len(out.data["X"])
            )
            f.write("{}\n".format(record))

            # Table
            data = np.transpose([out.data[k] for k in headers])
            for d in data:
                record = "".join("{:20.12e}".format(x) for x in d)
                f.write("{}{}\n".format("{:18}".format(""), record))



from __future__ import division, with_statement

from ...._common import block_to_format, get_label_length, open_file
from ...._exceptions import ReadError
from ...._helpers import FileIterator
from ._helpers import prune_nones_dict, prune_nones_list, read_record

__all__ = [
    "read",
]


oft_to_key = {
    "FOFT": "element_history",
    "COFT": "connection_history",
    "GOFT": "generator_history",
}


def read(filename, label_length=None, eos=None):
    """
    Read TOUGH input file.

    Parameters
    ----------
    filename : str
        Input file name.
    label_length : int or None, optional, default None
        Number of characters in cell labels.
    eos : str or None, optional, default None
        Equation of State.

    Returns
    -------
    dict
        TOUGH input parameters.

    """
    if not (label_length is None or isinstance(label_length, int)):
        raise TypeError()
    if isinstance(label_length, int) and not 5 <= label_length < 10:
        raise ValueError()

    with open_file(filename, "r") as f:
        out = read_buffer(f, label_length, eos)

    return out


def read_buffer(f, label_length, eos):
    """Read TOUGH input file."""
    from ._common import blocks

    parameters = {}

    # Title
    title = []
    while True:
        if len(title) >= 100:
            raise ValueError()

        line = f.readline().strip()

        if line[:5].upper() not in blocks:
            title.append(line)

        else:
            break

    if title:
        parameters["title"] = title[0] if len(title) == 1 else title

    f.seek(0)

    # Loop over blocks
    # Some blocks (INCON, INDOM, PARAM) need to rewind to previous line but tell and seek are disabled by next
    # See <https://stackoverflow.com/questions/22688505/is-there-a-way-to-go-back-when-reading-a-file-using-seek-and-calls-to-next>
    fiter = FileIterator(f)

    try:
        for line in fiter:
            if line.startswith("ROCKS"):
                parameters.update(_read_rocks(fiter))

            elif line.startswith("RPCAP"):
                rpcap = _read_rpcap(fiter)

                if "default" in parameters.keys():
                    parameters["default"].update(rpcap)

                else:
                    parameters["default"] = rpcap

            elif line.startswith("FLAC"):
                flac = _read_flac(fiter, parameters["rocks_order"])
                parameters["flac"] = flac["flac"]

                for k, v in flac["rocks"].items():
                    parameters["rocks"][k].update(v)

            elif line.startswith("CHEMP"):
                parameters.update(_read_chemp(fiter))

            elif line.startswith("NCGAS"):
                parameters.update(_read_ncgas(fiter))

            elif line.startswith("MULTI"):
                parameters.update(_read_multi(fiter))

            elif line.startswith("SOLVR"):
                parameters.update(_read_solvr(fiter))

            elif line.startswith("START"):
                parameters["start"] = True

            elif line.startswith("PARAM"):
                param = _read_param(fiter, eos)
                parameters["options"] = param["options"]
                parameters["extra_options"] = param["extra_options"]

                if "default" in parameters.keys():
                    parameters["default"].update(param["default"])

                else:
                    parameters["default"] = param["default"]

            elif line.startswith("SELEC"):
                parameters.update(_read_selec(fiter))

            elif line.startswith("INDOM"):
                indom = _read_indom(fiter, eos)

                for k, v in indom["rocks"].items():
                    parameters["rocks"][k].update(v)

            elif line.startswith("MOMOP"):
                parameters.update(_read_momop(fiter))

            elif line.startswith("TIMES"):
                parameters.update(_read_times(fiter))

            elif line.startswith("FOFT"):
                parameters.update(_read_oft(fiter, "FOFT", label_length))

            elif line.startswith("COFT"):
                parameters.update(_read_oft(fiter, "COFT", label_length))

            elif line.startswith("GOFT"):
                parameters.update(_read_oft(fiter, "GOFT", label_length))

            elif line.startswith("GENER"):
                parameters.update(_read_gener(fiter, label_length))

            elif line.startswith("DIFFU"):
                parameters.update(_read_diffu(fiter))

            elif line.startswith("OUTPU"):
                parameters.update(_read_outpu(fiter))

            elif line.startswith("ELEME"):
                parameters.update(_read_eleme(fiter, label_length))
                parameters["coordinates"] = False

            elif line.startswith("COORD"):
                coord = _read_coord(fiter)

                for k, v in zip(parameters["elements_order"], coord):
                    parameters["elements"][k]["center"] = v

                parameters["coordinates"] = True

            elif line.startswith("CONNE"):
                conne, flag = _read_conne(fiter, label_length)
                parameters.update(conne)

                if flag:
                    break

            elif line.startswith("INCON"):
                incon, flag = _read_incon(fiter, label_length, eos)
                parameters.update(incon)

                if flag:
                    break

            elif line.startswith("MESHM"):
                parameters.update(_read_meshm(fiter))

            elif line.startswith("NOVER"):
                parameters["nover"] = True

            elif line.startswith("ENDCY"):
                break

    except:
        raise ReadError("failed to parse line {}.".format(fiter.count))

    return parameters


def _read_rocks(f):
    """Read ROCKS block data."""
    fmt = block_to_format["ROCKS"]
    rocks = {"rocks": {}, "rocks_order": []}

    while True:
        line = f.next()

        if line.strip():
            # Record 1
            data = read_record(line, fmt[1])
            rock = data[0]
            rocks["rocks"][rock] = {
                "density": data[2],
                "porosity": data[3],
                "permeability": data[4] if len(set(data[4:7])) == 1 else data[4:7],
                "conductivity": data[7],
                "specific_heat": data[8],
            }

            nad = data[1]
            if nad is not None:
                # Record 2
                line = f.next()
                data = read_record(line, fmt[2])
                rocks["rocks"][rock].update(
                    {
                        "compressibility": data[0],
                        "expansivity": data[1],
                        "conductivity_dry": data[2],
                        "tortuosity": data[3],
                        "klinkenberg_parameter": data[4],
                        "distribution_coefficient_3": data[5],
                        "distribution_coefficient_4": data[6],
                    }
                )

            if nad and nad > 1:
                rocks["rocks"][rock].update(_read_rpcap(f))

            rocks["rocks_order"].append(rock)
        else:
            break

    rocks["rocks"] = {k: prune_nones_dict(v) for k, v in rocks["rocks"].items()}

    return rocks


def _read_rpcap(f):
    """Read RPCAP block data."""
    fmt = block_to_format["RPCAP"]
    rpcap = {}

    for key in ["relative_permeability", "capillarity"]:
        line = f.next()
        data = read_record(line, fmt)
        if data[0] is not None:
            rpcap[key] = {
                "id": data[0],
                "parameters": prune_nones_list(data[2:]),
            }

    return rpcap


def _read_flac(f, rocks_order):
    """Read FLAC block data."""
    fmt = block_to_format["FLAC"]
    flac = {"rocks": {}, "flac": {}}

    # Record 1
    line = f.next()
    data = read_record(line, fmt[1])
    flac["flac"]["creep"] = data[0]
    flac["flac"]["porosity_model"] = data[1]
    flac["flac"]["version"] = data[2]

    # Additional records
    for rock in rocks_order:
        flac["rocks"][rock] = {}

        line = f.next()
        data = read_record(line, fmt[2])
        flac["rocks"][rock]["permeability_model"] = {
            "id": data[0],
            "parameters": prune_nones_list(data[1:]),
        }

        line = f.next()
        data = read_record(line, fmt[3])
        flac["rocks"][rock]["equivalent_pore_pressure"] = {
            "id": data[0],
            "parameters": prune_nones_list(data[2:]),
        }

    flac["flac"] = prune_nones_dict(flac["flac"])

    return flac


def _read_chemp(f):
    """Read CHEMP block data."""
    fmt = block_to_format["CHEMP"]
    chemp = {"chemical_properties": {}}

    # Record 1
    line = f.next()
    n = read_record(line, fmt[1])[0]

    # Record 2
    for _ in range(n):
        tmp = {}

        line = f.next()
        data = read_record(line, fmt[2])
        chem = data[0]

        line = f.next()
        data = read_record(line, fmt[3])
        tmp["temperature_crit"] = data[0]
        tmp["pressure_crit"] = data[1]
        tmp["compressibility_crit"] = data[2]
        tmp["pitzer_factor"] = data[3]
        tmp["dipole_moment"] = data[4]

        line = f.next()
        data = read_record(line, fmt[3])
        tmp["boiling_point"] = data[0]
        tmp["vapor_pressure_a"] = data[1]
        tmp["vapor_pressure_b"] = data[2]
        tmp["vapor_pressure_c"] = data[3]
        tmp["vapor_pressure_d"] = data[4]

        line = f.next()
        data = read_record(line, fmt[3])
        tmp["molecular_weight"] = data[0]
        tmp["heat_capacity_a"] = data[1]
        tmp["heat_capacity_b"] = data[2]
        tmp["heat_capacity_c"] = data[3]
        tmp["heat_capacity_d"] = data[4]

        line = f.next()
        data = read_record(line, fmt[3])
        tmp["napl_density_ref"] = data[0]
        tmp["napl_temperature_ref"] = data[1]
        tmp["gas_diffusivity_ref"] = data[2]
        tmp["gas_temperature_ref"] = data[3]
        tmp["exponent"] = data[4]

        line = f.next()
        data = read_record(line, fmt[3])
        tmp["napl_viscosity_a"] = data[0]
        tmp["napl_viscosity_b"] = data[1]
        tmp["napl_viscosity_c"] = data[2]
        tmp["napl_viscosity_d"] = data[3]
        tmp["volume_crit"] = data[4]

        line = f.next()
        data = read_record(line, fmt[3])
        tmp["solubility_a"] = data[0]
        tmp["solubility_b"] = data[1]
        tmp["solubility_c"] = data[2]
        tmp["solubility_d"] = data[3]

        line = f.next()
        data = read_record(line, fmt[3])
        tmp["oc_coeff"] = data[0]
        tmp["oc_fraction"] = data[1]
        tmp["oc_decay"] = data[2]

        chemp["chemical_properties"][chem] = tmp

    return chemp


def _read_ncgas(f):
    """Read NCGAS block data."""
    fmt = block_to_format["NCGAS"]
    ncgas = {"non_condensible_gas": []}

    # Record 1
    line = f.next()
    n = read_record(line, fmt[1])[0]

    # Record 2
    for _ in range(n):
        line = f.next()
        data = read_record(line, fmt[2])
        ncgas["non_condensible_gas"].append(data[0])

    return ncgas


def _read_multi(f):
    """Read MULTI block data."""
    fmt = block_to_format["MULTI"]
    multi = {}

    line = f.next()
    data = read_record(line, fmt)
    multi["n_component"] = data[0]
    multi["isothermal"] = data[1] == data[0]
    multi["n_phase"] = data[2]
    multi["n_component_incon"] = data[4]

    return multi


def _read_solvr(f):
    """Read SOLVR block data."""
    fmt = block_to_format["SOLVR"]
    solvr = {}

    line = f.next()
    data = read_record(line, fmt)
    solvr["solver"] = {
        "method": data[0],
        "z_precond": data[2],
        "o_precond": data[4],
        "rel_iter_max": data[5],
        "eps": data[6],
    }

    return solvr


def _read_param(f, eos=None):
    """Read PARAM block data."""
    fmt = block_to_format["PARAM"]
    param = {}

    # Record 1
    line = f.next()
    data = read_record(line, fmt[1])
    param["options"] = {
        "n_iteration": data[0],
        "verbosity": data[1],
        "n_cycle": data[2],
        "n_second": data[3],
        "n_cycle_print": data[4],
        "temperature_dependence_gas": data[7],
        "effective_strength_vapor": data[8],
    }
    param["extra_options"] = {
        i + 1: int(x) for i, x in enumerate(data[5]) if x.isdigit()
    }

    # Record 2
    line = f.next()
    data = read_record(line, fmt[2])
    param["options"].update(
        {
            "t_ini": data[0],
            "t_max": data[1],
            "t_steps": data[2],
            "t_step_max": data[3],
            "gravity": data[5],
            "t_reduce_factor": data[6],
            "mesh_scale_factor": data[7],
        }
    )

    t_steps = int(data[2])
    if t_steps >= 0.0:
        param["options"]["t_steps"] = t_steps
    else:
        param["options"]["t_steps"] = []
        for _ in range(-t_steps):
            line = f.next()
            data = read_record(line, fmt[3])
            param["options"]["t_steps"] += prune_nones_list(data)
        if len(param["options"]["t_steps"]) == 1:
            param["options"]["t_steps"] = param["options"]["t_steps"][0]

    # Record 3
    line = f.next()
    data = read_record(line, fmt[4])
    param["options"].update(
        {
            "eps1": data[0],
            "eps2": data[1],
            "w_upstream": data[3],
            "w_newton": data[4],
            "derivative_factor": data[5],
        }
    )

    # Record 4 (TMVOC)
    if eos == "tmvoc":
        line = f.next()
        data = read_record(line, "5d")
        param["default"] = {"phase_composition": data[0]}

    # Record 4 and record 5 (EOS7R)
    line = f.next()
    data = read_record(line, fmt[5])

    i = f.tell()
    try:
        line = f.next()
        data += read_record(line, fmt[5])
    except ValueError:
        f.seek(i, increment=-1)

    if "default" not in param:
        param["default"] = {}

    if any(x is not None for x in data):
        data = prune_nones_list(data)
        param["default"]["initial_condition"] = data

    # Remove Nones
    param["options"] = prune_nones_dict(param["options"])
    param["extra_options"] = prune_nones_dict(param["extra_options"])

    return param


def _read_selec(f):
    """Read SELEC block data."""
    fmt = block_to_format["SELEC"]
    selec = {"selections": {}}

    line = f.next()
    data = read_record(line, fmt[1])
    selec["selections"]["integers"] = {k + 1: v for k, v in enumerate(data)}

    if selec["selections"]["integers"][1]:
        selec["selections"]["floats"] = []
        for _ in range(selec["selections"]["integers"][1]):
            line = f.next()
            data = read_record(line, fmt[2])
            selec["selections"]["floats"].append(prune_nones_list(data))

    selec["selections"]["integers"] = prune_nones_dict(selec["selections"]["integers"])
    if selec["selections"]["integers"][1] == 1:
        selec["selections"]["floats"] = selec["selections"]["floats"][0]

    return selec


def _read_indom(f, eos=None):
    """Read INDOM block data."""
    fmt = block_to_format["INDOM"]
    indom = {"rocks": {}}

    line = f.next()
    two_lines = True
    while True:
        if line.strip():
            # Record 1
            data = read_record(line, fmt[5])
            rock = data[0]
            phase_composition = data[1]  # TMVOC

            # Record 2
            line = f.next()
            data = read_record(line, fmt[0])

            # Record 3 (EOS7R)
            if two_lines:
                i = f.tell()
                line = f.next()

                if line.strip():
                    try:
                        data += read_record(line, fmt[0])
                    except ValueError:
                        two_lines = False
                        f.seek(i, increment=-1)
                else:
                    f.seek(i, increment=-1)

            data = prune_nones_list(data)
            indom["rocks"][rock] = {"initial_condition": data}

            if eos == "tmvoc":
                indom["rocks"][rock]["phase_composition"] = phase_composition

        else:
            break

        line = f.next()

    return indom


def _read_momop(f):
    """Read MOMOP block data."""
    fmt = block_to_format["MOMOP"]

    line = f.next()
    data = read_record(line, fmt)
    momop = {
        "more_options": {i + 1: int(x) for i, x in enumerate(data[0]) if x.isdigit()}
    }

    return momop


def _read_times(f):
    """Read TIMES block data."""
    fmt = block_to_format["TIMES"]
    times = {"times": []}

    # Record 1
    line = f.next()
    data = read_record(line, fmt[1])
    n_times = data[0]

    # Record 2
    while len(times["times"]) < n_times:
        line = f.next()
        data = read_record(line, fmt[2])
        times["times"] += prune_nones_list(data)

    return times


def _read_oft(f, oft, label_length):
    """Read FOFT, COFT and GOFT blocks data."""
    key = oft_to_key[oft]
    fmt = block_to_format[oft]
    history = {key: []}

    line = f.next()
    if not label_length:
        label_length = get_label_length(line[:9])

    while True:
        if line.strip():
            data = read_record(line, fmt[label_length])
            history[key].append(data[0])

        else:
            break

        line = f.next()

    return history


def _read_gener(f, label_length):
    """Read GENER block data."""
    fmt = block_to_format["GENER"]
    gener = {"generators": []}

    line = f.next()
    if not label_length:
        label_length = get_label_length(line[:9])

    while True:
        if line.strip():
            data = read_record(line, fmt[label_length])
            tmp = {
                "label": data[0],
                "name": data[1],
                "nseq": data[2],
                "nadd": data[3],
                "nads": data[4],
                "type": data[7],
                "layer_thickness": data[11],
            }

            ltab = data[5]
            if ltab and ltab > 1:
                itab = data[8]
                keys = ["times", "rates"]
                keys += ["specific_enthalpy"] if itab else []
                for key in keys:
                    table = []

                    while len(table) < ltab:
                        line = f.next()
                        data = read_record(line, fmt[0])
                        table += prune_nones_list(data)

                    tmp[key] = table

            else:
                tmp.update(
                    {"times": None, "rates": data[9], "specific_enthalpy": data[10],}
                )

            gener["generators"].append(tmp)

        else:
            break

        line = f.next()

    return {
        "generators": [prune_nones_dict(generator) for generator in gener["generators"]]
    }


def _read_diffu(f):
    """Read DIFFU block data."""
    fmt = block_to_format["DIFFU"]
    diffu = {"diffusion": []}

    while True:
        i = f.tell()
        line = f.next()

        if line.split():
            try:
                data = read_record(line, fmt)
                diffu["diffusion"].append(prune_nones_list(data))
            except ValueError:
                f.seek(i, increment=-1)
                break
        else:
            break

    return diffu


def _read_outpu(f):
    """Read OUTPU block data."""
    fmt = block_to_format["OUTPU"]
    outpu = {"output": {}}

    # Format
    line = f.next().strip()
    if line and not line.isdigit():
        outpu["output"]["format"] = line if line else None
        line = f.next().strip()

    # Variables
    if line.isdigit():
        num_vars = int(line)
        outpu["output"]["variables"] = []

        for _ in range(num_vars):
            line = f.next()
            data = read_record(line, fmt[3])
            name = data[0].lower()

            tmp = prune_nones_list(data[1:])
            options = None if len(tmp) == 0 else tmp[0] if len(tmp) == 1 else tmp

            outpu["output"]["variables"].append({"name": name})
            if options is not None:
                outpu["output"]["variables"][-1]["options"] = options

    return outpu


def _read_eleme(f, label_length):
    """Read ELEME block data."""
    fmt = block_to_format["ELEME"]
    eleme = {"elements": {}, "elements_order": []}

    line = f.next()
    if not label_length:
        label_length = get_label_length(line[:9])

    while True:
        if line.strip():
            data = read_record(line, fmt[label_length])
            label = data[0]
            rock = data[3].strip()
            eleme["elements"][label] = {
                "nseq": data[1],
                "nadd": data[2],
                "material": int(rock) if all(r.isdigit() for r in rock) else rock,
                "volume": data[4],
                "heat_exchange_area": data[5],
                "permeability_modifier": data[6],
                "center": data[7:10],
            }

            eleme["elements_order"].append(label)
        else:
            break

        line = f.next()

    eleme["elements"] = {k: prune_nones_dict(v) for k, v in eleme["elements"].items()}

    return eleme


def _read_coord(f):
    """Read COORD block data."""
    fmt = block_to_format["COORD"]
    coord = []

    line = f.next()
    while True:
        if line.strip():
            data = read_record(line, fmt)
            coord.append(data)
        else:
            break

        line = f.next()

    return coord


def _read_conne(f, label_length):
    """Read CONNE block data."""
    fmt = block_to_format["CONNE"]
    conne = {"connections": {}, "connections_order": []}

    line = f.next()
    if not label_length:
        label_length = get_label_length(line[:9])

    flag = False
    while True:
        if line.strip() and not line.startswith("+++"):
            data = read_record(line, fmt[label_length])
            label = data[0]
            conne["connections"][label] = {
                "nseq": data[1],
                "nadd": data[2:4],
                "permeability_direction": data[4],
                "nodal_distances": data[5:7],
                "interface_area": data[7],
                "gravity_cosine_angle": data[8],
                "radiant_emittance_factor": data[9],
            }

            conne["connections_order"].append(label)
        else:
            flag = line.startswith("+++")
            break

        line = f.next()

    conne["connections"] = {
        k: prune_nones_dict(v) for k, v in conne["connections"].items()
    }

    return conne, flag


def _read_incon(f, label_length, eos=None):
    """Read INCON block data."""
    fmt = block_to_format["INCON"]
    fmt2 = fmt[eos] if eos in fmt else fmt["default"]
    incon = {"initial_conditions": {}, "initial_conditions_order": []}

    line = f.next()
    if not label_length:
        label_length = get_label_length(line[:9])

    flag = False
    two_lines = True
    while True:
        if line.strip() and not line.startswith("+++"):
            # Record 1
            data = read_record(line, fmt2[label_length])
            label = data[0]
            incon["initial_conditions"][label] = {"porosity": data[3]}

            if eos == "tmvoc":
                incon["initial_conditions"][label]["phase_composition"] = data[4]

            else:
                userx = prune_nones_list(data[4:9])
                incon["initial_conditions"][label]["userx"] = userx if userx else None

            # Record 2
            line = f.next()
            data = read_record(line, fmt[0])

            # Record 3 (EOS7R)
            if two_lines:
                i = f.tell()
                line = f.next()

                if line.strip() and not line.startswith("+++"):
                    try:
                        data += read_record(line, fmt[0])
                    except ValueError:
                        two_lines = False
                        f.seek(i, increment=-1)
                else:
                    f.seek(i, increment=-1)

            incon["initial_conditions"][label]["values"] = prune_nones_list(data)
            incon["initial_conditions_order"].append(label)
        else:
            flag = line.startswith("+++")
            break

        line = f.next()

    incon["initial_conditions"] = {
        k: prune_nones_dict(v) for k, v in incon["initial_conditions"].items()
    }

    return incon, flag


def _read_meshm(f):
    """Read MESHM block data."""
    meshm = {"meshmaker": {}}
    fmt = block_to_format["MESHM"]

    # Mesh type
    line = f.next().strip()
    data = read_record(line, fmt[1])
    mesh_type = data[0].upper()

    # XYZ
    if mesh_type == "XYZ":
        meshm["meshmaker"]["type"] = mesh_type.lower()
        fmt = fmt["XYZ"]

        # Record 1
        line = f.next().strip()
        data = read_record(line, fmt[1])
        meshm["meshmaker"]["angle"] = data[0]

        # Record 2
        meshm["meshmaker"]["parameters"] = []

        while True:
            line = f.next()

            if line.strip():
                data = read_record(line, fmt[2])
                tmp = {
                    "type": data[0].lower(),
                    "n_increment": data[1],
                }

                if data[2]:
                    tmp["sizes"] = data[2]

                else:
                    sizes = []
                    while len(sizes) < tmp["n_increment"]:
                        line = f.next()
                        data = read_record(line, fmt[3])
                        sizes += prune_nones_list(data)

                    tmp["sizes"] = sizes[: tmp["n_increment"]]

                meshm["meshmaker"]["parameters"].append(tmp)

            else:
                break

    # RZ2D
    elif mesh_type in {"RZ2D", "RZ2DL"}:
        meshm["meshmaker"]["type"] = mesh_type.lower()
        fmt = fmt["RZ2D"]

        # Record 1
        meshm["meshmaker"]["parameters"] = []

        while True:
            line = f.next()

            if line.strip():
                data = read_record(line, fmt[1])
                data_type = data[0].upper()

                if data_type == "RADII":
                    line = f.next()
                    data = read_record(line, fmt["RADII"][1])
                    n = data[0]

                    radii = []
                    while len(radii) < n:
                        line = f.next()
                        data = read_record(line, fmt["RADII"][2])
                        radii += prune_nones_list(data)

                    tmp = {
                        "type": data_type.lower(),
                        "radii": radii[:n],
                    }
                    meshm["meshmaker"]["parameters"].append(tmp)

                elif data_type == "EQUID":
                    line = f.next()
                    data = read_record(line, fmt["EQUID"])

                    tmp = {
                        "type": data_type.lower(),
                        "n_increment": data[0],
                        "size": data[2],
                    }
                    meshm["meshmaker"]["parameters"].append(tmp)

                elif data_type == "LOGAR":
                    line = f.next()
                    data = read_record(line, fmt["LOGAR"])

                    tmp = {
                        "type": data_type.lower(),
                        "n_increment": data[0],
                        "radius": data[2],
                        "radius_ref": data[3],
                    }
                    meshm["meshmaker"]["parameters"].append(tmp)

                elif data_type == "LAYER":
                    # Record 1
                    line = f.next()
                    data = read_record(line, fmt["LAYER"][1])
                    n = data[0]

                    # Record 2
                    thicknesses = []
                    while len(thicknesses) < n:
                        line = f.next()
                        data = read_record(line, fmt["LAYER"][2])
                        thicknesses += prune_nones_list(data)

                    tmp = {"type": data_type.lower(), "thicknesses": thicknesses[:n]}
                    meshm["meshmaker"]["parameters"].append(tmp)

            else:
                break

    return meshm
